<?php

namespace Drupal\access_policy;

use Drupal\Core\Access\AccessResult;
use Drupal\Core\DependencyInjection\ContainerInjectionInterface;
use Drupal\Core\Entity\EntityInterface;
use Drupal\Core\Entity\EntityPublishedInterface;
use Drupal\Core\Session\AccountInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Access controller for content entities assigned an AccessPolicy.
 */
class AccessPolicyEntityAccessControlHandler implements ContainerInjectionInterface {

  /**
   * The access policy validator.
   *
   * @var \Drupal\access_policy\AccessPolicyValidator
   */
  protected $accessPolicyValidator;

  /**
   * The access policy selection service.
   *
   * @var \Drupal\access_policy\AccessPolicySelectionInterface
   */
  protected $accessPolicySelection;

  /**
   * Access policy information.
   *
   * @var \Drupal\access_policy\AccessPolicyInformation
   */
  protected $accessPolicyInformation;

  /**
   * Constructs a AccessPolicyEntityAccessControlHandler object.
   *
   * @param \Drupal\access_policy\AccessPolicyValidator $access_policy_validator
   *   The access policy validator.
   * @param \Drupal\access_policy\AccessPolicySelectionInterface $access_policy_selection
   *   The access policy selection handler.
   * @param \Drupal\access_policy\AccessPolicyInformation $policy_information
   *   The access policy information service.
   */
  public function __construct(AccessPolicyValidator $access_policy_validator, AccessPolicySelectionInterface $access_policy_selection, AccessPolicyInformation $policy_information) {
    $this->accessPolicyValidator = $access_policy_validator;
    $this->accessPolicySelection = $access_policy_selection;
    $this->accessPolicyInformation = $policy_information;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    return new static(
      $container->get('access_policy.validator'),
      $container->get('access_policy.selection'),
      $container->get('access_policy.information'),
    );
  }

  /**
   * Check access qualifications.
   *
   * @param \Drupal\Core\Entity\EntityInterface $entity
   *   The content entity.
   * @param \Drupal\Core\Session\AccountInterface $account
   *   The current user.
   * @param string $operation
   *   The current operation.
   *
   * @return \Drupal\Core\Access\AccessResultInterface
   *   The access result.
   */
  public function access(EntityInterface $entity, AccountInterface $account, string $operation) {

    if (!$this->accessPolicyInformation->hasEnabledForEntitiesOfEntityType($entity->getEntityType())) {
      return AccessResult::neutral();
    }

    $operation = $this->getCurrentOperation($entity, $operation);

    // If assigning an access policy confirm that the user has policies they
    // can assign first.
    if ($operation == 'manage access') {
      $access = $this->accessPolicySelection->validateAssignAccessPolicy($entity, $account);
      if (!$access) {
        return $this->accessResult($entity, $operation, $access);
      }
    }

    $access = $this->accessPolicyValidator->validate($entity, $account, $operation);

    return $this->accessResult($entity, $operation, $access);

  }

  /**
   * Final access result based on operation.
   *
   * @param \Drupal\Core\Entity\EntityInterface $entity
   *   The current entity.
   * @param string $operation
   *   The operation.
   * @param bool $allowed
   *   TRUE if allowed; FALSE otherwise.
   *
   * @return \Drupal\Core\Access\AccessResultInterface
   *   Access result.
   */
  private function accessResult(EntityInterface $entity, $operation, $allowed) {
    if (!$allowed) {
      return AccessResult::forbidden();
    }

    // The view unpublished operation requires an allowed result otherwise it
    // will return forbidden. All other operations work with neutral.
    switch ($operation) {
      case 'manage access':
        return AccessResult::allowed()->addCacheableDependency($entity);

      default;
        return AccessResult::neutral()->addCacheableDependency($entity);
    }
  }

  /**
   * Get the current operation.
   *
   * @param \Drupal\Core\Entity\EntityInterface $entity
   *   The entity.
   * @param string $operation
   *   The operation.
   *
   * @return string
   *   The current operation.
   */
  private function getCurrentOperation(EntityInterface $entity, $operation) {
    switch ($operation) {
      case 'view':
        if ($entity instanceof EntityPublishedInterface && !$entity->isPublished()) {
          return 'view unpublished';
        }
        break;

      case 'view revision':
      case 'revert revision':
      case 'delete revision':
        return 'view all revisions';
    }

    return $operation;
  }

}
