<?php

namespace Drupal\access_policy;

use Drupal\access_policy\Entity\AccessPolicyInterface;
use Drupal\Core\DependencyInjection\ContainerInjectionInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\StringTranslation\StringTranslationTrait;
use Drupal\access_policy\Entity\AccessPolicy;
use Drupal\user\EntityOwnerInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Defines dynamic permissions.
 *
 * @ingroup access_policy
 */
class AccessPolicyPermissions implements ContainerInjectionInterface {

  use StringTranslationTrait;

  /**
   * The entity type manager.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * AccessPolicyPermissions constructor.
   *
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   *   The entity type manager.
   */
  public function __construct(EntityTypeManagerInterface $entity_type_manager) {
    $this->entityTypeManager = $entity_type_manager;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    return new static($container->get('entity_type.manager'));
  }

  /**
   * Returns an array of Access Policy permissions.
   *
   * @return array
   *   The permissions.
   *
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   */
  public function entityPermissions() {
    $perms = [];
    // Generate permissions for all Access Policies based on their type.
    /** @var \Drupal\access_policy\Entity\AccessPolicyInterface $access_policy */
    foreach ($this->entityTypeManager->getStorage('access_policy')->loadMultiple() as $access_policy) {
      $perms = array_merge($perms, $this->createAccessPolicyPermissions($access_policy));
    }

    return $perms;
  }

  /**
   * Get the permission strings for a particular access policy.
   *
   * @param \Drupal\access_policy\Entity\AccessPolicyInterface $access_policy
   *   The access policy.
   *
   * @return array
   *   The access policy permissions.
   */
  public function getAccessPolicyPermissions(AccessPolicyInterface $access_policy) {
    $permissions = $this->createAccessPolicyPermissions($access_policy);
    if (!empty($permissions)) {
      return array_keys($permissions);
    }

    return [];
  }

  /**
   * Get the permissions for a specific access policy.
   *
   * @param \Drupal\access_policy\Entity\AccessPolicyInterface $access_policy
   *   The access policy.
   *
   * @return array[]
   *   Array of permissions.
   */
  protected function createAccessPolicyPermissions(AccessPolicyInterface $access_policy) {
    $perms = [];
    $entity_type = $this->entityTypeManager->getDefinition($access_policy->getTargetEntityTypeId());
    $entity_type_label = strtolower($entity_type->getLabel()->getUntranslatedString());
    $type = $access_policy->getTypePlugin();
    if ($type->supportsPermissions()) {
      $operations = $type->getPermissionOperations();
      foreach ($operations as $op) {
        switch ($op) {
          case 'view':
            $perms = array_merge($perms, $this->getViewPermission($access_policy, $entity_type_label));
            break;

          case 'update':
            $perms = array_merge($perms, $this->getEditPermission($access_policy, $entity_type_label));
            break;

          case 'delete':
            $perms = array_merge($perms, $this->getDeletePermission($access_policy, $entity_type_label));
            break;

          case 'view unpublished':
            // View own/any unpublished permission only exists in the
            // context of EntityOwnerInterface.
            if ($entity_type->entityClassImplements(EntityOwnerInterface::class)) {
              $perms = array_merge($perms, $this->getViewUnpublishedPermission($access_policy, $entity_type_label));
            }
            break;

          case 'view all revisions':
            if ($entity_type->isRevisionable() && $entity_type->hasLinkTemplate('revision')) {
              $perms = array_merge($perms, $this->getViewAllRevisionsPermission($access_policy, $entity_type_label));
            }
            break;

          case 'manage access':
            $perms = array_merge($perms, $this->getManageAccessPermission($access_policy, $entity_type_label));
            break;
        }
      }
    }

    // If the policy supports rules then create the bypass permission.
    if ($type->supportsRules()) {
      $perms = array_merge($perms, $this->getBypassAccessRulesPermission($access_policy, $entity_type_label));
    }

    // Assigning an access policy is controlled by permissions and selection
    // rules. Therefore, it can't be configured like other operations.
    $perms = array_merge($perms, $this->getAssignPermission($access_policy, $entity_type_label));

    // For each permission also append the dependency on the access policy.
    foreach ($perms as &$permission) {
      $permission['dependencies'] = $this->getAccessPolicyDependency($access_policy);
    }

    return $perms;
  }

  /**
   * Retrieves the view permission for a given entity type.
   *
   * @param \Drupal\access_policy\Entity\AccessPolicy $access_policy
   *   The Access Policy.
   * @param string $entity_type_label
   *   The entity type label.
   *
   * @return array
   *   The view permission.
   */
  private function getViewPermission(AccessPolicy $access_policy, $entity_type_label) {
    return [
      'view ' . $access_policy->id() . ' ' . $entity_type_label => [
        'title' => $this->t('%access_policy: View any %entity_type assigned this access policy', ['%access_policy' => $access_policy->label(), '%entity_type' => $entity_type_label]),
      ],
    ];
  }

  /**
   * Retrieves the view unpublished permission for a given entity type.
   *
   * @param \Drupal\access_policy\Entity\AccessPolicy $access_policy
   *   The Access Policy.
   * @param string $entity_type_label
   *   The entity type label.
   *
   * @return array
   *   The view unpublished permission.
   */
  private function getViewUnpublishedPermission(AccessPolicy $access_policy, $entity_type_label) {
    return [
      'view any ' . $access_policy->id() . ' unpublished ' . $entity_type_label => [
        'title' => $this->t('%access_policy: View any unpublished %entity_type assigned this access policy', ['%access_policy' => $access_policy->label(), '%entity_type' => $entity_type_label]),
      ],
    ];
  }

  /**
   * Retrieves the view all revisions permission for a given entity type.
   *
   * @param \Drupal\access_policy\Entity\AccessPolicy $access_policy
   *   The Access Policy.
   * @param string $entity_type_label
   *   The entity type label.
   *
   * @return array
   *   The view revisions permission.
   */
  private function getViewAllRevisionsPermission(AccessPolicy $access_policy, $entity_type_label) {
    return [
      'view all ' . $access_policy->id() . ' ' . $entity_type_label . ' revisions' => [
        'title' => $this->t('%access_policy: View all revisions of %entity_type assigned this access policy', ['%access_policy' => $access_policy->label(), '%entity_type' => $entity_type_label]),
        'description' => $this->t('To view the revisions, you still need permission to view all revisions.'),
      ],
    ];
  }

  /**
   * Retrieves the edit permission for a given entity type.
   *
   * @param \Drupal\access_policy\Entity\AccessPolicy $access_policy
   *   The Access Policy.
   * @param string $entity_type_label
   *   The entity type label.
   *
   * @return array
   *   The edit permission.
   */
  private function getEditPermission(AccessPolicy $access_policy, $entity_type_label) {
    return [
      'edit ' . $access_policy->id() . ' ' . $entity_type_label => [
        'title' => $this->t("%access_policy: Edt any %entity_type assigned this access policy", ['%access_policy' => $access_policy->label(), '%entity_type' => $entity_type_label]),
      ],
    ];
  }

  /**
   * Retrieves the delete permission for a given entity type.
   *
   * @param \Drupal\access_policy\Entity\AccessPolicy $access_policy
   *   The Access Policy.
   * @param string $entity_type_label
   *   The entity type label.
   *
   * @return array
   *   The listing permission.
   */
  private function getDeletePermission(AccessPolicy $access_policy, $entity_type_label) {
    return [
      'delete ' . $access_policy->id() . ' ' . $entity_type_label => [
        'title' => $this->t("%access_policy: Delete any %entity_type assigned this access policy", ['%access_policy' => $access_policy->label(), '%entity_type' => $entity_type_label]),
      ],
    ];
  }

  /**
   * Retrieves the assign permission for a given entity type.
   *
   * @param \Drupal\access_policy\Entity\AccessPolicy $access_policy
   *   The Access Policy.
   * @param string $entity_type_label
   *   The entity type label.
   *
   * @return array
   *   The listing permission.
   */
  private function getAssignPermission(AccessPolicy $access_policy, $entity_type_label) {
    return [
      'assign ' . $access_policy->id() . ' access policy' => [
        'title' => $this->t("%access_policy: Assign to any %entity_type", ['%access_policy' => $access_policy->label(), '%entity_type' => $entity_type_label]),
      ],
    ];
  }

  /**
   * Retrieves the manage access permission for a given entity type.
   *
   * Note that this was only added for the sake of completeness. It's currently
   * not being used in any of the policy types. For the most part, this is
   * satisfied by the 'set entity access policy' permission.
   *
   * @param \Drupal\access_policy\Entity\AccessPolicy $access_policy
   *   The Access Policy.
   * @param string $entity_type_label
   *   The entity type label.
   *
   * @return array
   *   The listing permission.
   */
  private function getManageAccessPermission(AccessPolicy $access_policy, $entity_type_label) {
    return [
      'manage ' . $access_policy->id() . ' access' => [
        'title' => $this->t("%access_policy: Manage the access policy for any %entity_type", ['%access_policy' => $access_policy->label(), '%entity_type' => $entity_type_label]),
      ],
    ];
  }

  /**
   * Retrieves the permission dependency config.
   *
   * This is primarily so that we can filter by permission on the manage
   * permissions tab.
   *
   * @param \Drupal\access_policy\Entity\AccessPolicy $access_policy
   *   The Access Policy.
   *
   * @return array
   *   The access policy dependency config.
   */
  private function getAccessPolicyDependency(AccessPolicy $access_policy) {
    return [
      $access_policy->getConfigDependencyKey() => [$access_policy->getConfigDependencyName()],
    ];
  }

  /**
   * Retrieves the assign permission for a given entity type.
   *
   * @param \Drupal\access_policy\Entity\AccessPolicy $access_policy
   *   The Access Policy.
   * @param string $entity_type_label
   *   The entity type label.
   *
   * @return array
   *   The listing permission.
   */
  private function getBypassAccessRulesPermission(AccessPolicy $access_policy, $entity_type_label) {
    return [
      'bypass ' . $access_policy->id() . ' access rules' => [
        'title' => $this->t("%access_policy: Bypass %entity_type access rule restrictions", ['%access_policy' => $access_policy->label(), '%entity_type' => $entity_type_label]),
        'description' => $this->t('View, edit and delete %entity_type regardless of access rule restrictions. Note that the user still needs permission for view/edit/delete any %entity_type assigned this access policy.', ['%entity_type' => $entity_type_label]),
      ],
    ];
  }

}
