<?php

namespace Drupal\access_policy;

use Drupal\Core\Entity\EntityInterface;
use Drupal\Core\Entity\EntityTypeInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Session\AccountInterface;
use Drupal\Core\StringTranslation\StringTranslationTrait;

/**
 * Handles the selection of access policies.
 */
class AccessPolicySelection implements AccessPolicySelectionInterface {

  use StringTranslationTrait;

  /**
   * Invalid set key.
   *
   * This can happen if an entity is assigned two or more policies that are
   * not part of a specific set.
   */
  const NULL_SET = 'null_set';

  /**
   * Define selection strategies.
   *
   * Selection strategies allow us to consolidate selection settings into
   * predefined sets. Later this could be consolidated into plugins if needed.
   *
   * @return array
   *   Array of selection strategies.
   */
  protected const SELECTION_STRATEGIES = [
    'automatic' => [
      'empty_allowed' => FALSE,
      'default_policy' => 'first_available',
      'on_entity_update' => 'change_policy',
      'policy_field' => 'disabled',
    ],
    'manual' => [
      'on_entity_update' => 'nothing',
      'policy_field' => 'enabled',
    ],
  ];

  /**
   * The default selection strategy.
   *
   * This is fallback if the strategy is not available in configuration.
   */
  protected const DEFAULT_STRATEGY = 'manual';

  /**
   * The entity type manager.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * Access policy information service.
   *
   * @var \Drupal\access_policy\AccessPolicyInformation
   */
  protected $accessPolicyInformation;

  /**
   * The selection rule plugin manager.
   *
   * @var \Drupal\access_policy\AccessPolicyHandlerManager
   */
  protected $selectionRuleManager;

  /**
   * Array of allowed access policies, keyed by entity id.
   *
   * @var array
   */
  protected $allowedPolicies = [];

  /**
   * Constructs a new AccessPolicySelection service.
   *
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   *   The entity type manager.
   * @param \Drupal\access_policy\AccessPolicyInformation $access_policy_information
   *   The access policy information service.
   * @param \Drupal\access_policy\AccessPolicyHandlerManager $selection_rule_manager
   *   The selection rule manager.
   */
  public function __construct(EntityTypeManagerInterface $entity_type_manager, AccessPolicyInformation $access_policy_information, AccessPolicyHandlerManager $selection_rule_manager) {
    $this->entityTypeManager = $entity_type_manager;
    $this->accessPolicyInformation = $access_policy_information;
    $this->selectionRuleManager = $selection_rule_manager;
  }

  /**
   * {@inheritdoc}
   */
  public function validateAssignAccessPolicy(EntityInterface $entity, $account) {

    if (!$this->hasAllowedAccessPolicies($entity, $account)) {
      return FALSE;
    }

    if (!$this->hasPermissionToCurrentPolicy($entity, $account)) {
      return FALSE;
    }

    return TRUE;
  }

  /**
   * Determine whether there are any allowed access policies.
   *
   * @param \Drupal\Core\Entity\EntityInterface $entity
   *   The current entity.
   * @param \Drupal\Core\Session\AccountInterface $account
   *   The current user.
   *
   * @return bool
   *   TRUE if the user can assign any policies; FALSE otherwise.
   *
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   */
  protected function hasAllowedAccessPolicies(EntityInterface $entity, AccountInterface $account) {
    $allowed_policies = $this->getAllowedAccessPolicies($entity, $account);
    if (empty($allowed_policies)) {
      return FALSE;
    }

    return TRUE;

  }

  /**
   * Determine whether the user has access to the current assigned policy.
   *
   * @param \Drupal\Core\Entity\EntityInterface $entity
   *   The entity.
   * @param \Drupal\Core\Session\AccountInterface $account
   *   The current user.
   *
   * @return bool
   *   TRUE if the user has access to the current policy; FALSE otherwise.
   */
  protected function hasPermissionToCurrentPolicy(EntityInterface $entity, AccountInterface $account) {
    // Compare allowed policies with the current policy. If not found, then do
    // not grant access.
    $current_policies = $entity->get('access_policy')->getValue();
    $current_policies = array_filter($current_policies, function ($item) {
      return (!empty($item['value']));
    });

    $policy_ids = array_map(function ($item) {
      return $item['value'];
    }, $current_policies);

    if (empty($policy_ids)) {
      return TRUE;
    }

    $policies = $this->entityTypeManager->getStorage('access_policy')->loadMultiple($policy_ids);
    foreach ($policies as $policy) {
      if (!$this->hasPermission($policy, $account)) {
        return FALSE;
      }
    }

    return TRUE;
  }

  /**
   * {@inheritdoc}
   */
  public function getAllowedAccessPolicies(EntityInterface $entity, AccountInterface $account) {
    if (isset($this->allowedPolicies[$entity->id()])) {
      return $this->allowedPolicies[$entity->id()];
    }

    $this->allowedPolicies[$entity->id()] = [];
    $policies = $this->getAccessPoliciesSorted($entity->getEntityType());
    foreach ($policies as $policy) {
      if ($this->isAllowed($policy, $entity, $account)) {
        $this->allowedPolicies[$entity->id()][] = $policy;
      }
    }

    return $this->allowedPolicies[$entity->id()];
  }

  /**
   * {@inheritdoc}
   */
  public function getApplicablePolicies(EntityInterface $entity, AccountInterface $account) {
    $policies = $this->getAccessPoliciesSorted($entity->getEntityType());
    if (!empty($policies)) {
      $candidates = [];
      foreach ($policies as $policy) {
        $is_allowed = $this->isAllowed($policy, $entity, $account);
        if ($is_allowed) {
          $selection_rules_valid = $this->selectionRulesValid($policy, $entity, $account);
          if ($selection_rules_valid) {
            // If this is the first policy and this is unique then return that
            // policy and exit.
            if (empty($candidates) && !$policy->isMultiple()) {
              return [$policy];
            }

            $candidates[$policy->getType()][] = $policy;
          }
        }
      }

      // Always return the first set of candidates.
      return reset($candidates);
    }

    return FALSE;
  }

  /**
   * {@inheritdoc}
   */
  public function getDefaultPolicy(EntityInterface $entity, AccountInterface $account) {
    // First check to see if the strategy has overridden this value.
    $default_value = $this->getStrategySetting('default_policy') ?? $this->getConfig()->get('default_policy');
    if ($default_value == 'first_available') {
      return $this->getApplicablePolicies($entity, $account);
    }

    return FALSE;
  }

  /**
   * Determine whether a user is allowed to assign an access policy.
   *
   * This checks whether the user has the permission or if the policy is
   * applicable.
   *
   * @param \Drupal\Core\Entity\EntityInterface $policy
   *   The access policy entity.
   * @param \Drupal\Core\Entity\EntityInterface $entity
   *   The current entity.
   * @param \Drupal\Core\Session\AccountInterface $account
   *   The current user.
   *
   * @return bool
   *   TRUE if the user can use the policy; FALSE otherwise.
   */
  protected function isAllowed(EntityInterface $policy, EntityInterface $entity, AccountInterface $account) {
    if ($this->isApplicable($policy, $entity) && $this->hasPermission($policy, $account)) {
      return TRUE;
    }

    return FALSE;
  }

  /**
   * Determine whether a user has the permission to assign an access policy.
   *
   * @param \Drupal\Core\Entity\EntityInterface $policy
   *   The access policy entity.
   * @param \Drupal\Core\Session\AccountInterface $account
   *   The current user.
   *
   * @return bool
   *   TRUE if the user can use the policy; FALSE otherwise.
   */
  protected function hasPermission(EntityInterface $policy, AccountInterface $account) {
    if (!$account->hasPermission('assign ' . $policy->id() . ' access policy')) {
      return FALSE;
    }
    return TRUE;
  }

  /**
   * Determine whether a policy is applicable for an entity.
   *
   * @param \Drupal\Core\Entity\EntityInterface $policy
   *   The access policy entity.
   * @param \Drupal\Core\Entity\EntityInterface $entity
   *   The current entity.
   *
   * @return bool
   *   TRUE if the policy is applicable; FALSE otherwise.
   *
   * @throws \Drupal\Component\Plugin\Exception\PluginException
   */
  protected function isApplicable(EntityInterface $policy, EntityInterface $entity) {
    // If the policy has selection rules, but they're not applicable to this
    // entity, then this policy is not allowed to be assigned.
    $has_handlers = !empty($policy->getHandlers('selection_rule'));
    $is_applicable = !empty($this->selectionRuleManager->getApplicableHandlers($policy, $entity));
    if ($has_handlers && !$is_applicable) {
      return FALSE;
    }

    return TRUE;
  }

  /**
   * Determine whether the selection rules are valid.
   *
   * @param \Drupal\Core\Entity\EntityInterface $policy
   *   The access policy.
   * @param \Drupal\Core\Entity\EntityInterface $entity
   *   The current entity.
   * @param \Drupal\Core\Session\AccountInterface $account
   *   The current user.
   *
   * @return bool
   *   TRUE if the selection rule are valid; FALSE otherwise.
   */
  protected function selectionRulesValid(EntityInterface $policy, EntityInterface $entity, AccountInterface $account) {
    $selection_rules = $this->selectionRuleManager->getHandlersFromPolicy($policy);
    if (!empty($selection_rules)) {
      $context = ExecutionContext::create('assign access policy', [
        'entity' => $entity,
        'user' => $account,
      ]);
      return $this->validateSelectionRules($selection_rules, $context, $policy->getSelectionRuleOperator());
    }

    return TRUE;
  }

  /**
   * Validate the selection rules.
   *
   * @param array $selection_rules
   *   Array of selection rules.
   * @param \Drupal\access_policy\ExecutionContextInterface $context
   *   The execution context.
   * @param string $operator
   *   The selection rule operator.
   *
   * @return bool
   *   TRUE if the selection rule passes; FALSE otherwise.
   */
  protected function validateSelectionRules(array $selection_rules, ExecutionContextInterface $context, $operator = 'AND') {
    $access = TRUE;
    foreach ($selection_rules as $plugin) {
      $plugin->setContext($context);
      // Compute the final value that will be validated against.
      $plugin->getValue();

      $entity = $context->get('entity');
      $account = $context->get('user');
      // @todo deprecate passing arguments to validate().
      $access = $plugin->validate($entity, $account);

      switch ($operator) {
        // If all rules are required then fail immediately if any return false.
        case 'AND':
          if (!$access) {
            return FALSE;
          }
          break;

        // If any rules are required then pass immediately if any return true.
        case 'OR':
          if ($access) {
            return TRUE;
          }
          break;
      }
    }

    return $access;
  }

  /**
   * Get access policies sorted by ascending weight.
   *
   * @param \Drupal\Core\Entity\EntityTypeInterface $entity_type
   *   The entity type.
   *
   * @return array
   *   Array of sorted access policies.
   *
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   */
  private function getAccessPoliciesSorted(EntityTypeInterface $entity_type) {
    $policies = $this->accessPolicyInformation->getEnabledForEntitiesOfEntityType($entity_type);
    $sorted = [];
    if (!empty($policies)) {
      foreach ($policies as $policy) {
        $sorted[$policy->id()] = $policy->getWeight();
      }
      asort($sorted);

      return $this->entityTypeManager->getStorage('access_policy')
        ->loadMultiple(array_keys($sorted));
    }

    return [];
  }

  /**
   * Determine whether empty access policies are allowed.
   *
   * @return bool
   *   TRUE if empty policies are allowed; FALSE otherwise.
   */
  public function emptyPoliciesAllowed() {
    // If the strategy has defined a value then use that instead.
    $empty_allowed = $this->getStrategySetting('allow_empty');
    if (isset($empty_allowed)) {
      return $empty_allowed;
    }

    if ($this->getConfig()->get('allow_empty')) {
      return TRUE;
    }
    return FALSE;
  }

  /**
   * Determine whether the policy should be updated on entity update.
   *
   * @return bool
   *   TRUE if the policy should be updated; FALSE otherwise.
   */
  public function changePolicyOnUpdate() {
    if ($this->getStrategySetting('on_entity_update') == 'change_policy') {
      return TRUE;
    }
    return FALSE;
  }

  /**
   * Determine whether the assignment field is enabled.
   *
   * @return bool
   *   TRUE if the policy should be updated; FALSE otherwise.
   */
  public function isPolicyFieldEnabled() {
    if ($this->getStrategySetting('policy_field') != 'disabled') {
      return TRUE;
    }
    return FALSE;
  }

  /**
   * Get the selection strategy setting value.
   *
   * @param string $key
   *   The strategy setting name.
   *
   * @return mixed
   *   The strategy setting value.
   */
  protected function getStrategySetting($key) {
    $strategy = $this->getStrategy();
    if (isset($strategy[$key])) {
      return $strategy[$key];
    }
    return NULL;
  }

  /**
   * Get the selection strategy.
   *
   * @return array
   *   The full strategy settings array.
   */
  protected function getStrategy() {
    $strategy = $this->getConfig()->get('selection_strategy') ?? self::DEFAULT_STRATEGY;
    return self::SELECTION_STRATEGIES[$strategy];
  }

  /**
   * {@inheritdoc}
   */
  public function getSelectionSetFromEntity(EntityInterface $entity) {
    $value = $entity->get('access_policy')->getValue();
    $policy_ids = array_map(function ($item) {
      return $item['value'];
    }, $value);
    $access_policies = $this->entityTypeManager->getStorage('access_policy')->loadMultiple($policy_ids);

    $assigned_sets = [];
    foreach ($access_policies as $policy) {
      $set = $policy->getSelectionSet();
      $assigned_sets = array_merge($assigned_sets, $set);
    }

    // No selection sets are assigned.
    if (empty($assigned_sets)) {
      return FALSE;
    }

    // If the count matches the same number of policies that that is the set
    // that policies are from.
    $total_policies = count($access_policies);
    $count_values = array_count_values($assigned_sets);
    $policy = array_search($total_policies, $count_values);

    // If no policy was found then return null set.
    if ($policy === FALSE) {
      return self::NULL_SET;
    }

    return $policy;
  }

  /**
   * {@inheritdoc}
   */
  public function getPoliciesInSelectionSet($name) {
    $policies = $this->entityTypeManager->getStorage('access_policy')->loadMultiple();
    return array_filter($policies, function ($policy) use ($name) {
      $selection_sets = $policy->getSelectionSet();
      if (in_array($name, $selection_sets)) {
        return TRUE;
      }
    });
  }

  /**
   * {@inheritdoc}
   */
  public function getSelectionSets() {
    return $this->getConfig()->get('selection_sets');
  }

  /**
   * {@inheritdoc}
   */
  public function getSelectionSet($name) {
    $sets = $this->getConfig()->get('selection_sets');
    if (isset($sets[$name])) {
      return $sets[$name];
    }

    return FALSE;
  }

  /**
   * Get the config settings.
   *
   * Note: This is currently a workaround to get tests to work. Currently,
   * config settings don't get passed in through dependency injection.
   *
   * @return \Drupal\Core\Config\ImmutableConfig
   *   A configuration object.
   */
  private function getConfig() {
    return \Drupal::service('config.factory')->get('access_policy.settings');
  }

}
