<?php

namespace Drupal\access_policy;

use Drupal\access_policy\Entity\AccessPolicyInterface;
use Drupal\Core\Entity\EntityInterface;

/**
 * Provides a trait for rendering arrays as labels.
 */
trait AccessPolicyTableTrait {

  /**
   * Get a list of policy type labels.
   *
   * @param \Drupal\Core\Entity\EntityInterface $entity
   *   The current entity.
   *
   * @return string
   *   The list of enabled entity types.
   *
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   */
  public function getPolicyTypeLabel(EntityInterface $entity) {
    $type = $entity->getTypePlugin();
    $definition = $type->getPluginDefinition();
    return $definition['label'];
  }

  /**
   * Get a list of selection sets.
   *
   * @param \Drupal\Core\Entity\EntityInterface $entity
   *   The current entity.
   * @param int $limit
   *   Set limit of how many labels to show.
   *
   * @return string
   *   The list of enabled entity types.
   *
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   */
  public function getSelectionSetLabels(EntityInterface $entity, $limit = NULL) {
    $selection_sets = $entity->getSelectionSet();
    $labels = [];
    foreach ($selection_sets as $name) {
      $set = $this->selectionManager->getSelectionSet($name);
      $labels[] = $set['label'];
    }

    return $this->renderLabels($labels, $limit);
  }

  /**
   * Render the labels.
   *
   * @param array $labels
   *   Array of labels.
   * @param int $limit
   *   Set limit of how many labels to show.
   *
   * @return \Drupal\Core\StringTranslation\TranslatableMarkup
   *   The rendered labels.
   */
  public function renderLabels(array $labels, $limit = NULL) {
    if (empty($labels)) {
      return $this->t('None');
    }

    $truncated = FALSE;
    $total = count($labels);
    if ($limit && count($labels) > $limit) {
      $labels = array_slice($labels, 0, $limit);
      $truncated = TRUE;
    }

    $string = implode(', ', $labels);
    if ($truncated) {
      $diff_count = $total - $limit;
      $string .= "… (" . $diff_count . " more)";
    }

    return $string;
  }

}
