<?php

namespace Drupal\access_policy;

use Drupal\Core\Database\Connection;
use Drupal\Core\Entity\EntityInterface;
use Drupal\access_policy\Entity\AccessPolicyInterface;
use Drupal\Core\Entity\EntityTypeInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\access_policy\Entity\ContentAccessPolicy;
use Drupal\access_policy\Entity\ContentAccessPolicyInterface;

/**
 * Service for handling Access Policy database content.
 *
 * This is added for legacy support ContentAccessPolicy storage.
 * This is based on entity storage which makes it difficult to support multiple
 * access policies on an entity. Moving away from entities to basic table
 * storage.
 *
 * @deprecated in access_policy:1.0.0-alpha8 and is removed from access_policy:1.0.0-beta1
 *   Use an EntityAccessPolicyStorage instead.
 *
 * @package Drupal\access_policy
 */
class ContentAccessPolicyStorage implements EntityAccessPolicyStorageInterface {

  /**
   * The entity type manager.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * Access policy information.
   *
   * @var \Drupal\access_policy\AccessPolicyInformation
   */
  protected $accessPolicyInformation;

  /**
   * Constructs a ContentAccessPolicyManager object.
   *
   * @param \Drupal\Core\Database\Connection $connection
   *   The database connection.
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   *   The entity type manager service.
   * @param \Drupal\access_policy\AccessPolicyInformation $policy_information
   *   Access policy information.
   */
  public function __construct(Connection $connection, EntityTypeManagerInterface $entity_type_manager, AccessPolicyInformation $policy_information) {
    $this->entityTypeManager = $entity_type_manager;
    $this->accessPolicyInformation = $policy_information;
  }

  /**
   * {@inheritdoc}
   */
  public function isUsed(AccessPolicyInterface $access_policy) {
    $result = $this->entityTypeManager->getStorage('content_access_policy')->getQuery()
      ->accessCheck(FALSE)
      ->condition('access_policy', $access_policy->id())
      ->range(0, 1)
      ->execute();

    return $result;
  }

  /**
   * {@inheritdoc}
   */
  public function getUsage(AccessPolicyInterface $access_policy) {
    $result = $this->entityTypeManager->getStorage('content_access_policy')->getQuery()
      ->accessCheck(FALSE)
      ->condition('access_policy', $access_policy->id())
      ->execute();

    return $result;
  }

  /**
   * {@inheritdoc}
   */
  public function read(EntityInterface $entity) {
    $content_access_policy = $this->entityTypeManager->getStorage('content_access_policy');

    $access_policy = $content_access_policy->getQuery()
      ->accessCheck(FALSE)
      ->condition('content_entity_type_id', $entity->getEntityTypeId())
      ->condition('content_entity_id', $entity->id())
      ->execute();

    if (empty($access_policy)) {
      return NULL;
    }

    /** @var \Drupal\access_policy\Entity\ContentAccessPolicyInterface $content_access_policy */
    $content_access_policy = $content_access_policy->load(key($access_policy));
    return [$content_access_policy->get('access_policy')->value];
  }

  /**
   * {@inheritdoc}
   */
  public function write(EntityInterface $entity, $access_policy) {
    /** @var \Drupal\Core\Entity\ContentEntityInterface $entity */
    $content_access_policy = ContentAccessPolicy::loadFromAccessControlledEntity($entity);
    /** @var \Drupal\Core\Entity\ContentEntityStorageInterface $storage */
    $storage = $this->entityTypeManager->getStorage('content_access_policy');

    if (!($content_access_policy instanceof ContentAccessPolicyInterface)) {
      $content_access_policy = $storage->create([
        'content_entity_type_id' => $entity->getEntityTypeId(),
        'content_entity_id' => $entity->id(),
      ]);
    }

    $content_access_policy->set('access_policy', $access_policy);
    ContentAccessPolicy::updateOrCreateFromEntity($content_access_policy);
  }

  /**
   * {@inheritdoc}
   */
  public function delete(EntityInterface $entity) {
    $content_access_policy = ContentAccessPolicy::loadFromAccessControlledEntity($entity);
    if ($content_access_policy) {
      $content_access_policy->delete();
    }
  }

  /**
   * {@inheritdoc}
   */
  public function getTable($entity_type) {
    return 'content_access_policy';
  }

  /**
   * {@inheritdoc}
   */
  public function getColumnName($key) {
    $column_map = [
      'entity_id' => 'content_entity_id',
      'value' => 'access_policy',
    ];

    if (isset($column_map[$key])) {
      return $column_map[$key];
    }

    return FALSE;
  }

  /**
   * {@inheritdoc}
   */
  public function isEmpty($entity_type) {
    $count = $this->entityTypeManager->getStorage('content_access_policy')
      ->getQuery()
      ->condition('content_entity_type_id', $entity_type)
      ->accessCheck(FALSE)
      ->count();
    if ($count == 0) {
      return TRUE;
    }
    return FALSE;
  }

  /**
   * Install the storage table if it doesn't exist.
   *
   * @param \Drupal\Core\Entity\EntityTypeInterface $entity_type
   *   The entity type id.
   *
   * @throws \Exception
   */
  public function installSchema(EntityTypeInterface $entity_type) {
    // Intentionally left blank.
  }

  /**
   * Uninstall the storage table.
   *
   * @param \Drupal\Core\Entity\EntityTypeInterface $entity_type
   *   The entity type id.
   *
   * @throws \Exception
   */
  public function uninstallSchema(EntityTypeInterface $entity_type) {
    // Intentionally left blank.
  }

}
