<?php

namespace Drupal\access_policy\Entity;

use Drupal\Core\Entity\ContentEntityBase;
use Drupal\Core\Entity\EntityInterface;
use Drupal\Core\Entity\EntityTypeInterface;
use Drupal\Core\Field\BaseFieldDefinition;
use Drupal\user\EntityOwnerTrait;

/**
 * Defines the Content access policy entity.
 *
 * @ContentEntityType(
 *   id = "content_access_policy",
 *   label = @Translation("Content access policy"),
 *   label_singular = @Translation("content access policy"),
 *   label_plural = @Translation("content access policies"),
 *   label_count = @PluralTranslation(
 *     singular = "@count content access policy",
 *     plural = "@count content access policy"
 *   ),
 *   handlers = {
 *     "storage_schema" = "Drupal\access_policy\ContentAccessPolicyStorageSchema",
 *     "views_data" = "\Drupal\views\EntityViewsData",
 *     "access" = "Drupal\access_policy\ContentAccessPolicyAccessControlHandler",
 *   },
 *   base_table = "content_access_policy",
 *   translatable = FALSE,
 *   internal = TRUE,
 *   entity_keys = {
 *     "id" = "id",
 *     "uuid" = "uuid",
 *   }
 * )
 *
 * @internal
 *   This entity is marked internal because it should not be used directly to
 *   alter the access policy of an entity. Instead, the computed
 *   accsss_policy field should be set on the entity directly.
 *
 * @deprecated in access_policy:1.0.0-alpha8 and is removed from access_policy:1.0.0-beta1
 *   Use an EntityAccessPolicyStorage instead.
 */
class ContentAccessPolicy extends ContentEntityBase implements ContentAccessPolicyInterface {

  use EntityOwnerTrait;

  /**
   * {@inheritdoc}
   */
  public static function baseFieldDefinitions(EntityTypeInterface $entity_type) {
    $fields = parent::baseFieldDefinitions($entity_type);

    $fields['access_policy'] = BaseFieldDefinition::create('string')
      ->setLabel(t('Access policy'))
      ->setDescription(t('The access policy of the the referenced content.'))
      ->setRequired(TRUE);

    $fields['content_entity_type_id'] = BaseFieldDefinition::create('string')
      ->setLabel(t('Content entity type ID'))
      ->setDescription(t('The ID of the content entity type this access policy is for.'))
      ->setRequired(TRUE)
      ->setSetting('max_length', EntityTypeInterface::ID_MAX_LENGTH);

    $fields['content_entity_id'] = BaseFieldDefinition::create('integer')
      ->setLabel(t('Content entity ID'))
      ->setDescription(t('The ID of the content entity this access policy is for.'))
      ->setRequired(TRUE);

    return $fields;
  }

  /**
   * Creates or updates an entity's access policy whilst saving that entity.
   *
   * @param \Drupal\access_policy\Entity\ContentAccessPolicyInterface $content_access_policy
   *   The content access policy entity content entity to create or save.
   *
   * @internal
   *   This method should only be called as a result of saving the related
   *   content entity.
   */
  public static function updateOrCreateFromEntity(ContentAccessPolicyInterface $content_access_policy) {
    $content_access_policy->realSave();
  }

  /**
   * Loads a content access policy entity.
   *
   * @param \Drupal\Core\Entity\EntityInterface $entity
   *   An access controlled entity object.
   *
   * @return \Drupal\access_policy\Entity\ContentAccessPolicyInterface
   *   The related content access policy or NULL if none could be found.
   *
   * @internal
   *   This method should only be called by code directly handling the
   *   ContentAccessPolicy entity objects.
   */
  public static function loadFromAccessControlledEntity(EntityInterface $entity) {
    $content_access_policy = NULL;
    $access_policy_info = \Drupal::service('access_policy.information');

    if ($access_policy_info->isAccessControlledEntity($entity)) {
      /** @var \Drupal\Core\Entity\ContentEntityInterface $entity */
      $storage = \Drupal::entityTypeManager()->getStorage('content_access_policy');

      $ids = $storage->getQuery()
        ->accessCheck(FALSE)
        ->condition('content_entity_type_id', $entity->getEntityTypeId())
        ->condition('content_entity_id', $entity->id())
        ->execute();

      if ($ids) {
        /** @var \Drupal\access_policy\Entity\ContentAccessPolicyInterface $content_access_policy */
        $content_access_policy = $storage->load(key($ids));
      }
    }

    return $content_access_policy;
  }

  /**
   * {@inheritdoc}
   */
  public function save() {
    $related_entity = \Drupal::entityTypeManager()
      ->getStorage($this->content_entity_type_id->value)
      ->load($this->content_entity_id->value);
    $related_entity->access_policy = $this->access_policy;
    return $related_entity->save();
  }

  /**
   * Saves an entity permanently.
   *
   * When saving existing entities, the entity is assumed to be complete,
   * partial updates of entities are not supported.
   *
   * @return int
   *   Either SAVED_NEW or SAVED_UPDATED, depending on the operation performed.
   *
   * @throws \Drupal\Core\Entity\EntityStorageException
   *   In case of failures an exception is thrown.
   */
  protected function realSave() {
    return parent::save();
  }

  /**
   * {@inheritdoc}
   */
  protected function getFieldsToSkipFromTranslationChangesCheck() {
    $field_names = parent::getFieldsToSkipFromTranslationChangesCheck();
    // We need to skip the parent entity revision ID, since that will always
    // change on every save, otherwise every translation would be marked as
    // affected regardless of actual changes.
    $field_names[] = 'content_entity_revision_id';
    return $field_names;
  }

}
