<?php

namespace Drupal\access_policy;

use Drupal\access_policy\Plugin\AccessPolicyType\AccessPolicyTypeInterface;
use Drupal\Core\DependencyInjection\ContainerInjectionInterface;
use Drupal\Core\Entity\EntityInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Form\FormStateInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Drupal\views\Views;

/**
 * Defines a class for reacting to entity events.
 *
 * @internal
 */
class EntityOperations implements ContainerInjectionInterface {

  use EntityAccessPolicyStorageTrait;

  /**
   * The entity type manager.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * The entity storage handler.
   *
   * @var \Drupal\access_policy\ContentAccessPolicyManager
   */
  protected $storage;

  /**
   * The access policy selection service.
   *
   * @var \Drupal\access_policy\AccessPolicySelectionInterface
   */
  protected $selection;

  /**
   * The access policy information service.
   *
   * @var \Drupal\access_policy\AccessPolicyInformation
   */
  protected $information;

  /**
   * Constructs a new EntityOperations object.
   *
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   *   The entity type manager.
   * @param \Drupal\access_policy\ContentAccessPolicyManager $storage
   *   The access policy content entity handler.
   * @param \Drupal\access_policy\AccessPolicySelectionInterface $selection
   *   The access policy selection service.
   * @param \Drupal\access_policy\AccessPolicyInformation $information
   *   The access policy information service.
   */
  public function __construct(EntityTypeManagerInterface $entity_type_manager, ContentAccessPolicyManager $storage, AccessPolicySelectionInterface $selection, AccessPolicyInformation $information) {
    $this->entityTypeManager = $entity_type_manager;
    $this->storage = $storage;
    $this->selection = $selection;
    $this->information = $information;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    return new static(
      $container->get('entity_type.manager'),
      $container->get('access_policy.content_policy_manager'),
      $container->get('access_policy.selection'),
      $container->get('access_policy.information'),
    );
  }

  /**
   * Assign an access policy to the entity.
   *
   * @param \Drupal\Core\Entity\EntityInterface $entity
   *   The entity.
   *
   * @see hook_entity_presave()
   */
  public function entityPresave(EntityInterface $entity) {
    if (!$this->storage->isPolicyUpdated($entity) && $this->information->hasEnabledForEntitiesOfEntityType($entity->getEntityType())) {
      $user = \Drupal::currentUser();
      $do_update = FALSE;
      if ($entity->isNew()) {
        $do_update = TRUE;
        $policy = $this->selection->getDefaultPolicy($entity, $user);
      }
      elseif ($this->selection->changePolicyOnUpdate()) {
        $do_update = TRUE;
        $policy = $this->selection->getApplicablePolicies($entity, $user);
      }

      if ($do_update) {
        // If no valid policy was found then remove any existing policy.
        if (empty($policy)) {
          $this->storage->clear($entity);
        }
        else {
          $this->storage->set($entity, $policy);
        }
      }
    }
  }

  /**
   * Add access rules and selection rules on initial save of access policy.
   *
   * @param \Drupal\Core\Entity\EntityInterface $entity
   *   The entity.
   *
   * @see hook_access_policy_presave()
   */
  public function accessPolicyPresave(EntityInterface $entity) {
    if ($entity->isNew()) {
      /** @var \Drupal\access_policy\Plugin\AccessPolicyType\AccessPolicyTypeInterface $type */
      $type = $entity->getTypePlugin();
      if ($type->hasTypeRules()) {
        $this->addRulesFromPolicyType($type, $entity);
      }
      if ($type->hasTypeSelectionRules()) {
        $this->addSelectionRulesFromPolicyType($type, $entity);
      }
    }
  }

  /**
   * Add access rules from the policy type (if any).
   *
   * @param \Drupal\access_policy\Plugin\AccessPolicyType\AccessPolicyTypeInterface $type
   *   The policy type object.
   * @param \Drupal\Core\Entity\EntityInterface $entity
   *   The current access policy.
   */
  private function addRulesFromPolicyType(AccessPolicyTypeInterface $type, EntityInterface $entity) {
    $rules = $type->getTypeRules();
    foreach ($rules as $rule) {
      $accessRuleManager = \Drupal::service('plugin.manager.access_policy.access_rule');
      $access_rule = $accessRuleManager->getHandler($entity->getTargetEntityTypeId(), $rule['id'], $rule);
      $entity->addHandler('access_rule', $access_rule);
    }
  }

  /**
   * Add selection rules from the policy type (if any).
   *
   * @param \Drupal\access_policy\Plugin\AccessPolicyType\AccessPolicyTypeInterface $type
   *   The policy type object.
   * @param \Drupal\Core\Entity\EntityInterface $entity
   *   The current access policy.
   */
  private function addSelectionRulesFromPolicyType(AccessPolicyTypeInterface $type, EntityInterface $entity) {
    $rules = $type->getTypeSelectionRules();
    foreach ($rules as $rule) {
      $accessRuleManager = \Drupal::service('plugin.manager.access_policy.selection_rule');
      $access_rule = $accessRuleManager->getHandler($entity->getTargetEntityTypeId(), $rule['id'], $rule);
      $entity->addHandler('selection_rule', $access_rule);
    }
  }

  /**
   * Setting the default access policy on the entity form.
   *
   * This is primarily for supporting constraint validation during initial
   * creation of the entity. The access policy actually gets saved as part of
   * hook_entity_presave.
   *
   * @see hook_form_alter();
   */
  public function entityFormAlter(&$form, FormStateInterface $form_state, $form_id) {
    $form_obj = $form_state->getFormObject();
    $entity = $form_obj->getEntity();

    $policies = $this->information->getEnabledForEntitiesOfEntityType($entity->getEntityType());
    // This only works if the policy doesn't have any selection rules. That's
    // because there is no way to know beforehand what policy is assigned
    // before the field has a value.
    $has_rules = FALSE;
    foreach ($policies as $policy) {
      $selection_rules = $policy->getHandlers('selection_rule');
      if (!empty($selection_rules)) {
        $has_rules = TRUE;
        break;
      }
    }

    if (!empty($policies) && !$has_rules && $entity->isNew()) {
      $default = $this->selection->getDefaultPolicy($entity, \Drupal::currentUser());
      if (!empty($default)) {
        $this->storage->set($entity, $default);
      }
    }
  }

  /**
   * Entity insert hook.
   *
   * @param \Drupal\Core\Entity\EntityInterface $entity
   *   The entity that was just saved.
   *
   * @see hook_entity_insert()
   */
  public function entityInsert(EntityInterface $entity) {
    if ($this->information->isAccessControlledEntity($entity)) {
      $this->updateOrCreateFromEntity($entity);
    }
  }

  /**
   * Entity update hook.
   *
   * @param \Drupal\Core\Entity\EntityInterface $entity
   *   The entity that was just saved.
   *
   * @see hook_entity_update()
   */
  public function entityUpdate(EntityInterface $entity) {
    if ($this->information->isAccessControlledEntity($entity)) {
      $this->updateOrCreateFromEntity($entity);
    }
  }

  /**
   * Creates or updates the access policy of an entity.
   *
   * @param \Drupal\Core\Entity\EntityInterface $entity
   *   The entity to update or create an access policy for.
   */
  protected function updateOrCreateFromEntity(EntityInterface $entity) {
    $access_policies = $entity->access_policy->getValue();

    $selected = array_map(function ($access_policy) {
      return $access_policy['value'];
    }, $access_policies);

    $storage = $this->getEntityPolicyStorage();
    $storage->write($entity, $selected);
  }

  /**
   * Entity delete hook.
   *
   * @param \Drupal\Core\Entity\EntityInterface $entity
   *   The entity being deleted.
   *
   * @see hook_entity_delete()
   */
  public function entityDelete(EntityInterface $entity) {
    if ($this->information->isAccessControlledEntity($entity)) {
      $storage = $this->getEntityPolicyStorage();
      $table = $storage->getTable($entity->getEntityTypeId());
      $column = $storage->getColumnName('entity_id');
      \Drupal::database()->delete($table)
        ->condition($column, $entity->id())
        ->execute();
    }
  }

  /**
   * Clear the caches after access policy update.
   *
   * @param \Drupal\Core\Entity\EntityInterface $entity
   *   The access policy entity.
   */
  public function accessPolicyUpdate(EntityInterface $entity) {
    // Clear field cache so extra field is added or removed.
    \Drupal::service('entity_field.manager')->clearCachedFieldDefinitions();
    // Clear the views data cache so the extra field is available in views.
    if (\Drupal::moduleHandler()->moduleExists('views')) {
      Views::viewsData()->clear();
    }
  }

  /**
   * Rebuild the routes.
   *
   * If this is the first access policy for a given entity type we need to
   * rebuild the routes so that the "Access" tab appears.
   *
   * @param \Drupal\Core\Entity\EntityInterface $entity
   *   The access policy entity.
   */
  public function accessPolicyInsert(EntityInterface $entity) {
    $existing_policies = $this->entityTypeManager->getStorage('access_policy')
      ->loadByProperties(['target_entity_type_id' => $entity->getTargetEntityTypeId()]);
    // If this is the first policy for a given entity type then clear the
    // routes and register the access policy data.
    if (count($existing_policies) == 1) {
      \Drupal::service('router.builder')->setRebuildNeeded();
      \Drupal::service('access_policy.access_policy_data')->clear();
    }

    // Clear field cache so extra field is added or removed.
    \Drupal::service('entity_field.manager')->clearCachedFieldDefinitions();
    // Clear the views data cache so the extra field is available in views.
    if (\Drupal::moduleHandler()->moduleExists('views')) {
      Views::viewsData()->clear();
    }
  }

}
