<?php

namespace Drupal\access_policy;

use Drupal\Core\DependencyInjection\ContainerInjectionInterface;
use Drupal\Core\Entity\EntityFieldManagerInterface;
use Drupal\Core\Entity\EntityTypeInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Field\FieldDefinitionInterface;
use Drupal\Core\StringTranslation\StringTranslationTrait;
use Drupal\field\FieldConfigInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Class to manage and lazy load cached access policy data.
 */
class FieldMatchAccessPolicyData implements ContainerInjectionInterface {

  use StringTranslationTrait;

  /**
   * The entity type manager.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * The entity field manager.
   *
   * @var \Drupal\Core\Entity\EntityFieldManagerInterface
   */
  protected $entityFieldManager;

  /**
   * The access policy information service.
   *
   * @var \Drupal\access_policy\AccessPolicyInformation
   */
  protected $accessPolicyInfo;

  /**
   * Static cache of field definitions.
   *
   * @var array
   */
  protected $fieldDefinitions = [];

  /**
   * Constructs the FieldMatchAccessPolicyData access policy data set.
   *
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   *   The entity type manager.
   * @param \Drupal\Core\Entity\EntityFieldManagerInterface $entity_field_manager
   *   The entity field manager.
   * @param \Drupal\access_policy\AccessPolicyInformation $access_policy_info
   *   The access policy information service.
   */
  public function __construct(EntityTypeManagerInterface $entity_type_manager, EntityFieldManagerInterface $entity_field_manager, AccessPolicyInformation $access_policy_info) {
    $this->entityTypeManager = $entity_type_manager;
    $this->entityFieldManager = $entity_field_manager;
    $this->accessPolicyInfo = $access_policy_info;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    return new static(
      $container->get('entity_type.manager'),
      $container->get('entity_field.manager'),
      $container->get('access_policy.information'),
    );
  }

  /**
   * Defines the default access policy data.
   *
   * @see hook_access_policy_data()
   */
  public function accessPolicyData() {
    $data = [];
    $this->createDerivedAccessRules($data);
    return $data;
  }

  /**
   * Create derived access rules.
   *
   * @param array $data
   *   The current access rule data.
   */
  protected function createDerivedAccessRules(array &$data) {
    $target_types = $this->entityTypeManager->getDefinitions();
    $entity_types = $this->accessPolicyInfo->getAllEnabledEntityTypes();
    foreach ($target_types as $target_type) {
      if ($this->userHasEntityReferenceField($target_type->id())) {
        foreach ($entity_types as $entity_type) {
          $field_definitions = $this->getFieldDefinitions($entity_type, ['entity_reference']);
          foreach ($field_definitions as $definition) {
            $storage = $definition->getFieldStorageDefinition();
            $is_entity_ref_field = ($definition instanceof FieldConfigInterface && $storage->getSetting('target_type') == $target_type->id());
            if ($is_entity_ref_field) {
              $data[$entity_type->id()]['match_' . $definition->getName()] = $this->createFieldMatchEntityReferenceAccessRule($entity_type, $definition);
            }
          }
        }
      }
    }

    if ($this->userHasListField()) {
      $entity_types = $this->accessPolicyInfo->getAllEnabledEntityTypes();
      foreach ($entity_types as $entity_type) {
        $list_fields = [
          'list_string',
          'list_float',
          'list_integer',
        ];
        $field_definitions = $this->getFieldDefinitions($entity_type, $list_fields);
        foreach ($field_definitions as $definition) {
          $data[$entity_type->id()]['match_' . $definition->getName()] = $this->createFieldMatchListAccessRule($entity_type, $definition);
        }
      }
    }

    $entity_types = $this->accessPolicyInfo->getAllEnabledEntityTypes();
    foreach ($entity_types as $entity_type) {

      $has_author_field = $entity_type->hasKey('owner') && $this->userHasEntityReferenceField('user');
      if ($has_author_field) {
        $data[$entity_type->id()]['author_reference'] = $this->createAuthorReferenceAccessRule($entity_type);
      }

      if ($this->userHasEntityReferenceField($entity_type->id())) {
        $data[$entity_type->id()]['entity_reference'] = $this->createCurrentEntityReferenceAccessRule($entity_type);
      }

      $field_definitions = $this->getFieldDefinitions($entity_type, ['entity_reference']);
      foreach ($field_definitions as $definition) {
        $field_storage = $definition->getFieldStorageDefinition();
        $is_user_ref_field = ($definition instanceof FieldConfigInterface && $field_storage->getSetting('target_type') == 'user');
        if ($is_user_ref_field) {
          $data[$entity_type->id()][$definition->getName() . '_user_reference'] = $this->createUserReferenceAccessRule($entity_type, $definition);
        }
      }
    }
  }

  /**
   * Create a field match entity reference access rule.
   *
   * @param \Drupal\Core\Entity\EntityTypeInterface $entity_type
   *   The entity type.
   * @param \Drupal\Core\Field\FieldDefinitionInterface $definition
   *   The field definition.
   *
   * @return array
   *   The access rule data definition.
   *
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   */
  protected function createFieldMatchListAccessRule(EntityTypeInterface $entity_type, FieldDefinitionInterface $definition) {
    return [
      'label' => $this->t('Field match: @entity_field', ['@entity_field' => $definition->getLabel()]),
      'description' => $this->t('Compare values between the @entity_type and current user.', [
        '@entity_type' => $entity_type->getSingularLabel(),
      ]),
      'plugin_id' => 'entity_field_list',
      'entity_type' => $entity_type->id(),
      'field' => $definition->getName(),
      'argument' => [
        'plugin' => 'current_user',
        'field_type' => $definition->getType(),
      ],
      'widget' => 'entity_field',
    ];
  }

  /**
   * Create a field match entity reference access rule.
   *
   * @param \Drupal\Core\Entity\EntityTypeInterface $entity_type
   *   The entity type.
   * @param \Drupal\Core\Field\FieldDefinitionInterface $definition
   *   The field definition.
   *
   * @return array
   *   The access rule data definition.
   *
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   */
  protected function createFieldMatchEntityReferenceAccessRule(EntityTypeInterface $entity_type, FieldDefinitionInterface $definition) {
    $field_storage = $definition->getFieldStorageDefinition();
    $target_type_id = $field_storage->getSetting('target_type');
    $target_type = $this->entityTypeManager->getDefinition($target_type_id);
    return [
      'label' => $this->t('Field match: @entity_field', ['@entity_field' => $definition->getLabel()]),
      'description' => $this->t('Compare @target_type between the @entity_type and current user.', [
        '@target_type' => $target_type->getPluralLabel(),
        '@entity_type' => $entity_type->getSingularLabel(),
      ]),
      'plugin_id' => 'entity_field_entity_reference',
      'entity_type' => $entity_type->id(),
      'field' => $definition->getName(),
      'argument' => [
        'plugin' => 'current_user',
        'field_type' => 'entity_reference',
      ],
      'widget' => 'entity_field',
    ];
  }

  /**
   * Create the user reference access rule.
   *
   * @param \Drupal\Core\Entity\EntityTypeInterface $entity_type
   *   The entity type.
   * @param \Drupal\Core\Field\FieldDefinitionInterface $definition
   *   The field definition.
   *
   * @return array
   *   The access rule data definition.
   */
  protected function createUserReferenceAccessRule(EntityTypeInterface $entity_type, FieldDefinitionInterface $definition) {
    return [
      'label' => $this->t('@field_name field has reference to current user', [
        '@field_name' => $definition->getLabel(),
        '@entity_type' => $entity_type->getSingularLabel(),
      ]),
      'description' => $this->t('The @entity_type has a reference to the current user.', ['@entity_type' => $entity_type->getSingularLabel()]),
      'plugin_id' => 'entity_field_standard',
      'operator' => '=',
      'entity_type' => $entity_type->id(),
      'field' => $definition->getName(),
      'argument' => [
        'plugin' => 'current_user',
        'field' => 'uid',
      ],
      'widget' => 'entity_field',
    ];
  }

  /**
   * Create author reference access rule.
   *
   * @param \Drupal\Core\Entity\EntityTypeInterface $entity_type
   *   The entity type.
   *
   * @return array
   *   The access rule data definition.
   */
  protected function createAuthorReferenceAccessRule(EntityTypeInterface $entity_type) {
    $uid_field = $entity_type->getKey('owner');
    return [
      'label' => $this->t('Current user: Has reference to original author'),
      'description' => $this->t('The current user has a reference to the original author of the @entity_type.', ['@entity_type' => $entity_type->getSingularLabel()]),
      'plugin_id' => 'entity_field_entity_reference',
      'operator' => 'in',
      'entity_type' => $entity_type->id(),
      'field' => $uid_field,
      'argument' => [
        'plugin' => 'current_user',
        'field_type' => 'entity_reference',
      ],
    ];
  }

  /**
   * Create the entity reference access rule.
   *
   * @param \Drupal\Core\Entity\EntityTypeInterface $entity_type
   *   The entity type.
   *
   * @return array
   *   The access rule data definition.
   */
  public function createCurrentEntityReferenceAccessRule(EntityTypeInterface $entity_type) {
    $id_field = $entity_type->getKey('id');
    return [
      'label' => $this->t('Current user: Has reference to this @entity_type', ['@entity_type' => $entity_type->getSingularLabel()]),
      'description' => $this->t('The current user has a reference to this @entity_type.', ['@entity_type' => $entity_type->getSingularLabel()]),
      'plugin_id' => 'entity_field_standard',
      'operator' => '=',
      'entity_type' => $entity_type->id(),
      'field' => $id_field,
      'argument' => [
        'plugin' => 'current_user',
        'field_type' => 'entity_reference',
      ],
    ];
  }

  /**
   * Determine whether the user has an list field.
   *
   * @return bool
   *   TRUE if the user has a list field; FALSE otherwise.
   */
  protected function userHasListField() {
    $list_fields = [
      'list_string',
      'list_integer',
      'list_float',
    ];
    $fields = $this->entityFieldManager->getFieldDefinitions('user', 'user');
    foreach ($fields as $field) {
      if (in_array($field->getType(), $list_fields)) {
        return TRUE;
      }
    }
    return FALSE;
  }

  /**
   * Determine whether the user has an entity reference field.
   *
   * @param string $target_type
   *   The target type (e.g node, taxonomy term etc.).
   *
   * @return bool
   *   TRUE if the user has an entity reference field; FALSE otherwise.
   */
  protected function userHasEntityReferenceField($target_type) {
    $fields = $this->entityFieldManager->getFieldDefinitions('user', 'user');
    foreach ($fields as $field) {
      if ($field->getType() == 'entity_reference') {
        $storage = $field->getFieldStorageDefinition();
        if ($storage->getSetting('target_type') == $target_type) {
          return TRUE;
        }
      }
    }
    return FALSE;
  }

  /**
   * Get the field definition for a specific field type.
   *
   * Note that it only returns a definition for one supported bundle.
   *
   * @param \Drupal\Core\Entity\EntityTypeInterface $entity_type
   *   The entity type.
   * @param array $field_types
   *   The field types array.
   *
   * @return \Drupal\Core\Field\FieldDefinitionInterface[]
   *   Array of field definitions.
   */
  protected function getFieldDefinitions(EntityTypeInterface $entity_type, array $field_types) {
    $field_definitions = [];
    foreach ($field_types as $field_type) {
      $field_map = $this->entityFieldManager->getFieldMapByFieldType($field_type);
      $fields = $field_map[$entity_type->id()] ?? [];

      foreach ($fields as $field_name => $data) {
        $bundles = $data['bundles'];
        foreach ($bundles as $bundle) {
          if (empty($this->fieldDefinitions[$bundle])) {
            $this->fieldDefinitions[$bundle] = $this->entityFieldManager->getFieldDefinitions($entity_type->id(), $bundle);
          }
          if (isset($this->fieldDefinitions[$bundle][$field_name]) && !isset($field_definitions[$field_name])) {
            $field_definitions[$field_name] = $this->fieldDefinitions[$bundle][$field_name];
          }
        }
      }
    }
    return $field_definitions;
  }

}
