<?php

namespace Drupal\access_policy\Plugin\AccessRuleWidget;

use Drupal\Core\Entity\Entity\EntityFormDisplay;
use Drupal\Core\Entity\EntityInterface;
use Drupal\Core\Form\FormStateInterface;

/**
 * Embedded entity field access rule widget.
 *
 * Renders  the entity field on the access tab.
 *
 * @AccessRuleWidget(
 *   id = "entity_field",
 *   label = @Translation("Entity field"),
 *   description = @Translation("Shows the field on the access tab.")
 * )
 */
class EntityField extends AccessRuleWidgetBase {

  /**
   * {@inheritdoc}
   */
  public function settingsForm(array $form, FormStateInterface $form_state) {
    // Get the widget options based on field type.
    $field_type = $this->definition->getFieldType();
    $widgetPluginManager = \Drupal::service('plugin.manager.field.widget');
    $options = $widgetPluginManager->getOptions($field_type);

    $form['field_widget'] = [
      '#title' => $this->t('Field widget display'),
      '#type' => 'select',
      '#default_value' => $this->getWidgetSetting('field_widget'),
      '#options' => $options,
    ];
    return $form;
  }

  /**
   * {@inheritdoc}
   */
  public function buildForm($form, $form_state) {
    $field_name = $this->definition->getFieldName();
    $entity = $this->getEntity();

    $items = $entity->get($field_name);
    $items->filterEmptyItems();

    $widget = $this->getFieldWidget($field_name, $form, $form_state);
    $widget_form = $widget->form($items, $form, $form_state);

    $is_disabled = FALSE;
    if ($entity->isTranslatable() && $this->isNotTranslatableField($entity, $field_name)) {
      $url = $entity->getUntranslated()->toUrl('access-policy-form')->toString();
      \Drupal::messenger()->addWarning($this->t('Fields that apply to all languages are disabled to avoid conflicting changes. <a href=":url">Edit them on the original language access form</a>.', [':url' => $url]));
      $is_disabled = TRUE;
    }

    // Disable the widget if the access policy selection field is also disabled.
    $selection = \Drupal::service('access_policy.selection');
    if (!$selection->isPolicyFieldEnabled()) {
      $is_disabled = TRUE;
    }

    if ($is_disabled) {
      $widget_form['widget']['#disabled'] = TRUE;
    }

    return $widget_form;
  }

  /**
   * Determine whether the field is translateable.
   *
   * @param \Drupal\Core\Entity\EntityInterface $entity
   *   The current entity.
   * @param string $field_name
   *   The current field.
   *
   * @return bool
   *   TRUE if not translateable; FALSE otherwise.
   */
  private function isNotTranslatableField(EntityInterface $entity, $field_name) {
    $items = $entity->get($field_name);
    $definition = $items->getFieldDefinition();
    if (!$entity->language()->isDefault() && !$definition->isTranslatable()) {
      return TRUE;
    }

    return FALSE;
  }

  /**
   * {@inheritdoc}
   */
  public function submitForm(array &$form, FormStateInterface $form_state) {
    $field_name = $this->definition->getFieldName();
    $value = $form_state->getValue($field_name);

    // Normally this is handled by WidgetBase however it also makes a lot of
    // other assumptions about the form which are not true in this case.
    if (isset($value['add_more'])) {
      unset($value['add_more']);
    }

    $widget = $this->getFieldWidget($field_name, $form, $form_state);
    $value = $widget->massageFormValues($value, $form, $form_state);

    $entity = $this->getEntity();
    $entity->set($field_name, $value);
    $entity->save();
  }

  /**
   * Get the field widget for the field.
   *
   * @param string $field_name
   *   The field name.
   * @param array $form
   *   The current form.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The currnet form state.
   *
   * @return \Drupal\Core\Field\PluginSettingsInterface|null
   *   A widget or formatter plugin.
   */
  protected function getFieldWidget($field_name, array &$form, FormStateInterface $form_state) {
    $entity = $this->getEntity();
    $new_form_display = EntityFormDisplay::create([
      'targetEntityType' => $entity->getEntityTypeId(),
      'bundle' => $entity->bundle(),
      'mode' => 'default',
    ]);

    $options = [];
    if (!empty($this->getWidgetSetting('field_widget'))) {
      $options['type'] = $this->getWidgetSetting('field_widget');
    }
    $new_form_display->setComponent($field_name, $options);

    $widget = $new_form_display->getRenderer($field_name);
    $new_form_display->buildForm($entity, $form, $form_state);

    return $widget;
  }

}
