<?php

namespace Drupal\access_policy\Plugin\Field;

use Drupal\access_policy\EntityAccessPolicyStorageTrait;
use Drupal\Core\Entity\ContentEntityInterface;
use Drupal\Core\Field\FieldItemList;
use Drupal\Core\TypedData\ComputedItemListTrait;

/**
 * A computed field that provides a content entity's access policy.
 *
 * It links content entities to an access policy entity via a content access
 * policy content entity.
 */
class AccessPolicyFieldItemList extends FieldItemList {

  use EntityAccessPolicyStorageTrait;
  use ComputedItemListTrait {
    get as traitGet;
  }

  /**
   * {@inheritdoc}
   */
  protected function computeValue() {
    $access_policy = $this->getAccessPolicyId();
    // Do not store NULL values, in the case where an entity does not have an
    // access policy associated with it, we do not create list items for
    // the computed field.
    if ($access_policy) {
      foreach ($access_policy as $index => $policy) {
        $this->list[$index] = $this->createItem($index, $policy);
      }
    }
  }

  /**
   * Gets the access policy ID linked to a content entity.
   *
   * @return array
   *   The array of access policy IDs linked to a content entity.
   */
  protected function getAccessPolicyId() {
    $entity = $this->getEntity();

    /** @var \Drupal\access_policy\AccessPolicyInformation $access_policy_info */
    $access_policy_info = \Drupal::service('access_policy.information');
    if (!$access_policy_info->hasEnabledForEntitiesOfEntityType($entity->getEntityType())) {
      return NULL;
    }

    if (!$entity->isNew() && $content_access_policy = $this->loadContentAccessPolicy($entity)) {
      return $content_access_policy;
    }

  }

  /**
   * Load the content access policy associated with an entity.
   *
   * @param \Drupal\Core\Entity\ContentEntityInterface $entity
   *   The entity the access policy entity will be loaded from.
   *
   * @return string|null
   *   The content access policy or FALSE if none exists.
   */
  protected function loadContentAccessPolicy(ContentEntityInterface $entity) {
    $storage = $this->getEntityPolicyStorage();
    return $storage->read($entity);
  }

  /**
   * {@inheritdoc}
   */
  public function get($index) {
    return $this->traitGet($index);
  }

  /**
   * {@inheritdoc}
   */
  public function setValue($values, $notify = TRUE) {
    parent::setValue($values, $notify);
    $this->valueComputed = TRUE;
  }

  /**
   * {@inheritdoc}
   */
  public function generateSampleItems($count = 1) {
    // No sample items generated since the starting access policy is always
    // computed based on the default value.
  }

}
