<?php

namespace Drupal\Tests\access_policy\Functional;

use Drupal\access_policy\Entity\AccessPolicy;
use Drupal\access_policy\EntityAccessPolicyStorageTrait;
use Drupal\field\Entity\FieldConfig;
use Drupal\field\Entity\FieldStorageConfig;
use Drupal\node\Entity\NodeType;
use Drupal\taxonomy\Entity\Vocabulary;
use Drupal\Tests\taxonomy\Traits\TaxonomyTestTrait;

/**
 * Ensures that access policy entity permissions work correctly.
 *
 * @group access_policy
 */
class AccessPolicyEntityAssignTest extends AccessPolicyTestBase {

  use EntityAccessPolicyStorageTrait;
  use TaxonomyTestTrait;

  /**
   * The tags vocabulary.
   *
   * @var \Drupal\Core\Entity\EntityInterface
   */
  protected $vocabulary;

  /**
   * Modules to enable.
   *
   * @var array
   */
  protected static $modules = [
    'access_policy',
    'access_policy_test',
    'filter',
    'node',
    'taxonomy',
    'datetime',
  ];

  /**
   * {@inheritdoc}
   */
  protected $defaultTheme = 'stark';

  /**
   * Set to TRUE to strict check all configuration saved.
   *
   * @var bool
   *
   * @see \Drupal\Core\Config\Development\ConfigSchemaChecker
   */
  protected $strictConfigSchema = FALSE;

  /**
   * {@inheritdoc}
   */
  public function setUp(): void {
    parent::setup();

    // Create a tags vocabulary.
    $this->vocabulary = Vocabulary::create([
      'name' => 'tags',
      'vid' => 'tags',
    ]);
    $this->vocabulary->save();

    $field_storage = FieldStorageConfig::create([
      'field_name' => 'field_tags',
      'entity_type' => 'node',
      'type' => 'entity_reference',
      'cardinality' => -1,
      'settings' => [
        'target_type' => 'taxonomy_term',
      ],
    ]);
    $field_storage->save();

    $field = FieldConfig::create([
      'field_storage' => $field_storage,
      'field_name' => 'field_tags',
      'bundle' => 'page',
      'translatable' => FALSE,
    ]);
    $field->save();

    $field_storage = FieldStorageConfig::create([
      'field_name' => 'field_tags',
      'entity_type' => 'user',
      'type' => 'entity_reference',
      'cardinality' => -1,
      'settings' => [
        'target_type' => 'taxonomy_term',
      ],
    ]);
    $field_storage->save();

    $field = FieldConfig::create([
      'field_storage' => $field_storage,
      'field_name' => 'field_tags',
      'bundle' => 'user',
      'translatable' => FALSE,
    ]);
    $field->save();

    AccessPolicy::create([
      'id' => 'test_policy',
      'label' => 'Test policy',
      'access_rules' => [],
      'target_entity_type_id' => 'node',
      'weight' => 0,
    ])->save();

    AccessPolicy::create([
      'id' => 'test_policy_2',
      'label' => 'Test policy 2',
      'access_rules' => [],
      'target_entity_type_id' => 'node',
      'weight' => -10,
    ])->save();

    \Drupal::configFactory()->getEditable('access_policy.settings')
      ->set('default_policy', 'first_available')
      ->save();

    \Drupal::service('router.builder')->rebuildIfNeeded();
  }

  /**
   * Tests accessing the access tab for entity without access policy.
   */
  public function testAccessEntityWithoutAccessPolicy() {
    // Create an article content type that does not have access policy
    // enabled.
    NodeType::create([
      'type' => 'article',
      'name' => 'Article',
    ])->save();

    $node = $this->drupalCreateNode([
      'type' => 'article',
      'status' => 1,
    ]);

    $web_user = $this->drupalCreateUser([
      'set entity access policy',
      'edit any article content',
    ]);
    $this->drupalLogin($web_user);

    // They should be able to view the node like normal.
    $this->drupalGet('node/' . $node->id());
    $this->assertSession()->statusCodeEquals(200);

    // They should be able to edit the node like normal.
    $this->drupalGet('node/' . $node->id() . '/edit');
    $this->assertSession()->statusCodeEquals(200);

    $this->drupalGet('node/' . $node->id() . '/access');
    $this->assertSession()->statusCodeEquals(403);
  }

  /**
   * Tests assigning a new policy only restricted by permissions.
   */
  public function testAssignNewPolicy() {
    \Drupal::configFactory()->getEditable('access_policy.settings')
      ->set('default_policy', '')
      ->save();

    // Having edit any page content and edit test_policy content should give
    // you access to the content.
    $web_user = $this->drupalCreateUser([
      'set entity access policy',
      'assign test_policy access policy',
      'assign test_policy_2 access policy',
    ]);
    $this->drupalLogin($web_user);

    $node = $this->drupalCreateNode([
      'type' => 'page',
      'status' => 1,
      'uid' => $web_user->id(),
    ]);

    $this->drupalGet('node/' . $node->id() . '/access');
    $this->assertSession()->statusCodeEquals(200);
    $this->assertSession()->optionExists('access_policy', 'test_policy');
    $this->assertSession()->optionExists('access_policy', 'test_policy_2');
    $this->assertSession()->fieldValueEquals('access_policy', '');

    $edit = [
      'access_policy' => 'test_policy_2',
    ];
    $this->submitForm($edit, 'Save');
    $this->assertSession()->fieldValueEquals('access_policy', 'test_policy_2');

    // Confirm that the content access policy was created when assigned to the
    // node.
    $policy = $this->getEntityPolicyStorage()->read($node);
    $this->assertNotEmpty($policy, 'The content access policy entity should have been created');

    $node = $this->drupalGetNodeByTitle($node->label());
    $policy = $this->contentAccessPolicyManager->getAccessPolicy($node);
    $policy = reset($policy);
    $this->assertEquals('test_policy_2', $policy->id(), 'The access policy was incorrect.');

    // Delete the node and confirm that the content access policy entity was
    // also deleted.
    $node->delete();
    $policy = $this->getEntityPolicyStorage()->read($node);
    $this->assertEmpty($policy, 'The content access policy entity should have been deleted');
  }

  /**
   * Tests assigning a policy with selections rules on a node with no policy.
   */
  public function testAssignNewPolicyWithSelectionRule() {
    \Drupal::configFactory()->getEditable('access_policy.settings')
      ->set('default_policy', '')
      ->save();

    // Article is not applicable.
    NodeType::create([
      'type' => 'article',
      'name' => 'Article',
    ])->save();

    $access_policy = $this->entityTypeManager->getStorage('access_policy')->load('test_policy');

    $selection_rule = $this->selectionRuleManager->getHandler('node', 'field_text');
    $access_policy->addHandler('selection_rule', $selection_rule);
    $access_policy->save();

    // Node has no access policy.
    $node_page = $this->drupalCreateNode([
      'type' => 'page',
      'status' => 1,
    ]);

    $node_article = $this->drupalCreateNode([
      'type' => 'article',
      'status' => 1,
    ]);

    $web_user = $this->drupalCreateUser([
      'set entity access policy',
      'assign test_policy access policy',
      'assign test_policy_2 access policy',
    ]);

    $this->drupalLogin($web_user);

    $this->drupalGet('node/' . $node_page->id() . '/access');
    $this->assertSession()->statusCodeEquals(200);
    $this->assertSession()->optionExists('access_policy', 'test_policy');
    $this->assertSession()->optionExists('access_policy', 'test_policy_2');

    $this->drupalGet('node/' . $node_article->id() . '/access');
    $this->assertSession()->statusCodeEquals(200);
    $this->assertSession()->optionNotExists('access_policy', 'test_policy');
    $this->assertSession()->optionExists('access_policy', 'test_policy_2');
  }

  /**
   * Tests assigning an access policy with an access rule.
   */
  public function testAssignNewPolicyWithAccessRule() {
    \Drupal::configFactory()->getEditable('access_policy.settings')
      ->set('default_policy', '')
      ->save();
    // Assign test_policy access policy to the node.
    $access_policy = $this->entityTypeManager->getStorage('access_policy')->load('test_policy');
    $handler = $this->accessRuleManager->getHandler('node', 'match_field_tags');
    $handler->setSettings([
      'operator' => 'in',
      'widget' => [
        'show' => TRUE,
      ],
      'value' => [
        'field' => 'field_tags',
      ],
    ]);

    $access_policy->addHandler('access_rule', $handler);
    $access_policy->save();

    // Test entity reference fields.
    $term_1 = $this->createTerm($this->vocabulary, [
      'title' => 'Term 1',
    ]);
    $term_1->save();

    // Assigning the inapplicable selection rule.
    $node = $this->drupalCreateNode([
      'type' => 'page',
      'status' => 1,
    ]);

    $web_user = $this->drupalCreateUser([
      'set entity access policy',
      'assign test_policy access policy',
      'assign test_policy_2 access policy',
    ]);
    $web_user->set('field_tags', [
      $term_1,
    ])->save();

    $this->drupalLogin($web_user);

    $this->drupalGet('node/' . $node->id() . '/access');
    $this->assertSession()->statusCodeEquals(200);
    $this->assertSession()->fieldValueEquals('access_policy', '');
    $this->assertSession()->optionExists('access_policy', 'test_policy');
    $this->assertSession()->optionExists('access_policy', 'test_policy_2');

    // Assign the access policy with access rule so that we can set the tag.
    $this->assignAccessPolicy($node, 'test_policy');
    $this->drupalGet('node/' . $node->id() . '/access');
    $this->assertSession()->statusCodeEquals(200);
    $this->assertSession()->fieldValueEquals('access_policy', 'test_policy');

    $term_input = $term_1->label() . ' (' . $term_1->id() . ')';
    $edit = [
      'field_tags[0][target_id]' => $term_input,
    ];
    $this->submitForm($edit, 'Save');
    $this->assertSession()->fieldValueEquals('field_tags[0][target_id]', $term_input);
    $this->assertSession()->pageTextContains('Access has been changed to Test policy.');
    $node = $this->drupalGetNodeByTitle($node->label());
    $this->assertEquals($term_1->id(), $node->get('field_tags')->target_id);
  }

  /**
   * Tests changing the current policy to another access policy.
   */
  public function testChangePolicy() {
    // Page content type has the field so it should show the policy.
    $node = $this->drupalCreateNode([
      'type' => 'page',
      'status' => 1,
    ]);
    $this->assignAccessPolicy($node, 'test_policy');

    $web_user = $this->drupalCreateUser([
      'set entity access policy',
      'assign test_policy access policy',
      'assign test_policy_2 access policy',
    ]);

    $this->drupalLogin($web_user);
    $this->drupalGet('node/' . $node->id() . '/access');
    $this->assertSession()->statusCodeEquals(200);
    $this->assertSession()->fieldValueEquals('access_policy', 'test_policy');

    $edit = [
      'access_policy' => 'test_policy_2',
    ];
    $this->submitForm($edit, 'Save');
    $this->assertSession()->fieldValueEquals('access_policy', 'test_policy_2');

    $node = $this->drupalGetNodeByTitle($node->label());
    $policy = $this->contentAccessPolicyManager->getAccessPolicy($node);
    $policy = reset($policy);
    $this->assertEquals('test_policy_2', $policy->id(), 'The access policy was incorrect.');

    // This user does not have permission to assign any policies. It should
    // trigger a 403.
    $web_user = $this->drupalCreateUser([
      'set entity access policy',
    ]);
    $this->drupalLogin($web_user);
    $this->drupalGet('node/' . $node->id() . '/access');
    $this->assertSession()->statusCodeEquals(403);

    // This user does not have permission to assign test_policy_2. It should
    // trigger a 403.
    $web_user = $this->drupalCreateUser([
      'set entity access policy',
      'assign test_policy access policy',
    ]);
    $this->drupalLogin($web_user);
    $this->drupalGet('node/' . $node->id() . '/access');
    $this->assertSession()->statusCodeEquals(403);

    $web_user = $this->drupalCreateUser([
      'set entity access policy',
      'assign test_policy_2 access policy',
    ]);
    $this->drupalLogin($web_user);
    $this->drupalGet('node/' . $node->id() . '/access');
    $this->assertSession()->statusCodeEquals(200);
  }

  /**
   * Tests changing an access policy with a selection rule.
   */
  public function testChangePolicyWithSelectionRule() {

    $access_policy = $this->entityTypeManager->getStorage('access_policy')->load('test_policy_2');
    $selection_rule = $this->selectionRuleManager->getHandler('node', 'field_text');
    $access_policy->addHandler('selection_rule', $selection_rule);
    $access_policy->save();

    // Page content type has the field so it should show the policy.
    $node = $this->drupalCreateNode([
      'type' => 'page',
      'status' => 1,
    ]);
    $this->assignAccessPolicy($node, 'test_policy');

    $web_user = $this->drupalCreateUser([
      'set entity access policy',
      'assign test_policy access policy',
      'assign test_policy_2 access policy',
    ]);

    $this->drupalLogin($web_user);
    $this->drupalGet('node/' . $node->id() . '/access');
    $this->assertSession()->statusCodeEquals(200);
    $this->assertSession()->fieldValueEquals('access_policy', 'test_policy');

    $edit = [
      'access_policy' => 'test_policy_2',
    ];
    $this->submitForm($edit, 'Save');
    $this->assertSession()->fieldValueEquals('access_policy', 'test_policy_2');

    $node = $this->drupalGetNodeByTitle($node->label());
    $policy = $this->contentAccessPolicyManager->getAccessPolicy($node);
    $policy = reset($policy);
    $this->assertEquals('test_policy_2', $policy->id(), 'The access policy was incorrect.');
  }

  /**
   * Tests changing the access policy with an inapplicable selection rule.
   *
   * This is testing an edge case where a selection rule may have been
   * applicable once but after assignment.
   */
  public function testChangePolicyWithInapplicableSelectionRule() {
    // Article does not have field_text so it should not be applicable.
    NodeType::create([
      'type' => 'article',
      'name' => 'Article',
    ])->save();

    // Assign test_policy access policy to the node.
    $access_policy = $this->entityTypeManager->getStorage('access_policy')->load('test_policy');
    $selection_rule = $this->selectionRuleManager->getHandler('node', 'field_text');
    $access_policy->addHandler('selection_rule', $selection_rule);
    $access_policy->save();

    // Assigning the inapplicable selection rule.
    $node_article = $this->drupalCreateNode([
      'type' => 'article',
      'status' => 1,
    ]);
    $this->assignAccessPolicy($node_article, 'test_policy');

    // This does not have any access policy and therefore should not show
    // test_policy.
    $node_article_2 = $this->drupalCreateNode([
      'type' => 'article',
      'status' => 1,
    ]);

    $web_user = $this->drupalCreateUser([
      'set entity access policy',
      'assign test_policy access policy',
      'assign test_policy_2 access policy',
    ]);

    $this->drupalLogin($web_user);

    $this->drupalGet('node/' . $node_article->id() . '/access');
    $this->assertSession()->statusCodeEquals(200);
    $this->assertSession()->fieldValueEquals('access_policy', 'test_policy');
    $this->assertSession()->optionExists('access_policy', 'test_policy_2');

    // Change the policy, test_policy should then be removed from the select
    // list.
    $edit = [
      'access_policy' => 'test_policy_2',
    ];
    $this->submitForm($edit, 'Save');
    $this->assertSession()->fieldValueEquals('access_policy', 'test_policy_2');
    $this->assertSession()->fieldValueEquals('access_policy', 'test_policy_2');
    $this->assertSession()->optionNotExists('access_policy', 'test_policy');

    $this->drupalGet('node/' . $node_article_2->id() . '/access');
    $this->assertSession()->statusCodeEquals(200);
    $this->assertSession()->optionNotExists('access_policy', 'test_policy');
    $this->assertSession()->optionExists('access_policy', 'test_policy_2');
  }

  /**
   * Tests accessing the access tab when it has the type selection rule.
   */
  public function testChangeAccessPolicyWithTypeSelectionRule() {
    NodeType::create([
      'type' => 'article',
      'name' => 'Article',
    ])->save();

    $access_policy = $this->entityTypeManager->getStorage('access_policy')->load('test_policy');
    $selection_rule = $this->selectionRuleManager->getHandler('node', 'type');
    $selection_rule->setSettings([
      'value' => [
        'page' => 'page',
      ],
      'operator' => 'or',
    ]);
    $access_policy->addHandler('selection_rule', $selection_rule);
    $access_policy->save();

    $web_user = $this->drupalCreateUser([
      'set entity access policy',
      'assign test_policy access policy',
      'assign test_policy_2 access policy',
    ]);

    $this->drupalLogin($web_user);

    // Page content type has the field so it should show the policy.
    $node = $this->drupalCreateNode([
      'type' => 'page',
      'status' => 1,
    ]);
    $this->drupalGet('node/' . $node->id() . '/access');
    $this->assertSession()->statusCodeEquals(200);
    $this->assertSession()->optionExists('access_policy', 'Test policy');
    $this->assertSession()->optionExists('access_policy', 'Test policy 2');

    // Article content type does not support Test policy, so it should not
    // appear.
    $node = $this->drupalCreateNode([
      'type' => 'article',
      'status' => 1,
    ]);
    $this->drupalGet('node/' . $node->id() . '/access');
    $this->assertSession()->statusCodeEquals(200);
    $this->assertSession()->optionNotExists('access_policy', 'Test policy');
    $this->assertSession()->optionExists('access_policy', 'Test policy 2');
  }

  /**
   * Tests changing a policy that has an access rule.
   */
  public function testChangePolicyWithAccessRule() {
    // Assign test_policy access policy to the node.
    $access_policy = $this->entityTypeManager->getStorage('access_policy')->load('test_policy');
    $handler = $this->accessRuleManager->getHandler('node', 'match_field_tags');
    $handler->setSettings([
      'operator' => 'in',
      'value' => [
        'field' => 'field_tags',
      ],
    ]);

    $access_policy->addHandler('access_rule', $handler);
    $access_policy->save();

    // Test entity reference fields.
    $term_1 = $this->createTerm($this->vocabulary, [
      'title' => 'Term 1',
    ]);
    $term_1->save();

    $term_2 = $this->createTerm($this->vocabulary, [
      'title' => 'Term 1',
    ]);
    $term_2->save();

    // Assigning the inapplicable selection rule.
    $node = $this->drupalCreateNode([
      'type' => 'page',
      'status' => 1,
      'field_tags' => [
        $term_1,
      ],
    ]);
    $this->assignAccessPolicy($node, 'test_policy_2');

    $web_user = $this->drupalCreateUser([
      'set entity access policy',
      'assign test_policy access policy',
      'assign test_policy_2 access policy',
    ]);
    $web_user->set('field_tags', [
      $term_1,
    ])->save();

    $this->drupalLogin($web_user);

    $this->drupalGet('node/' . $node->id() . '/access');
    $this->assertSession()->statusCodeEquals(200);
    $this->assertSession()->fieldValueEquals('access_policy', 'test_policy_2');
    $this->assertSession()->optionExists('access_policy', 'test_policy');

    // Change the policy to one with an access rule.
    $edit = [
      'access_policy' => 'test_policy',
    ];
    $this->submitForm($edit, 'Save');
    $this->assertSession()->statusCodeEquals(200);
    $this->assertSession()->fieldValueEquals('access_policy', 'test_policy');
    $this->assertSession()->optionExists('access_policy', 'test_policy_2');

    // This user has the same permissions but different tag, therefore they
    // shouldn't have access.
    $web_user = $this->drupalCreateUser([
      'set entity access policy',
      'assign test_policy access policy',
      'assign test_policy_2 access policy',
    ]);
    $web_user->set('field_tags', [
      $term_2,
    ])->save();
    $this->drupalLogin($web_user);
    $this->drupalGet('node/' . $node->id() . '/access');
    $this->assertSession()->statusCodeEquals(403);
  }

  /**
   * Test that the default access policy gets saved to the node when created.
   */
  public function testDefaultAccessPolicyOnNodeCreate() {
    $web_user = $this->drupalCreateUser([
      'assign test_policy access policy',
    ]);
    $this->drupalLogin($web_user);
    $node = $this->drupalCreateNode([
      'type' => 'page',
      'status' => 1,
    ]);
    $node->save();

    $this->assertEquals('test_policy', $node->get('access_policy')->value, 'The access policy should be test_policy');
  }

  /**
   * Tests removing an access policy from a node via a form.
   */
  public function testSetEmptyAccessPolicyOnEntity() {
    // Allow the access policy to be empty.
    \Drupal::configFactory()->getEditable('access_policy.settings')
      ->set('allow_empty', TRUE)
      ->save();

    // Create a published page. Under normal circumstances this will always
    // be visible.
    $node = $this->drupalCreateNode([
      'type' => 'page',
      'status' => 1,
    ]);

    // Assign test_policy access policy to the node.
    $access_policy = \Drupal::entityTypeManager()->getStorage('access_policy')->load('test_policy');
    $this->contentAccessPolicyManager->assign($node, $access_policy);

    // Having edit any page content and edit test_policy content should give
    // you access to the content.
    $web_user = $this->drupalCreateUser([
      'set entity access policy',
      'assign test_policy access policy',
    ]);
    $this->drupalLogin($web_user);

    $this->drupalGet('node/' . $node->id() . '/access');
    $this->assertSession()->statusCodeEquals(200);

    // Assert that the "- None -" option exists in the select list.
    $this->assertSession()->pageTextContains('- None -');

    // Remove the access policy and then assert that the access policy is
    // marked as removed and it won't revert to the default value again.
    $edit = [
      'access_policy' => '',
    ];
    $this->submitForm($edit, 'Save');

    $node = $this->drupalGetNodeByTitle($node->label());
    $empty = $this->contentAccessPolicyManager->isEmpty($node);

    $this->assertTrue($empty, 'The access policy should have been removed.');
    $this->assertSession()->fieldValueEquals('access_policy', '');
  }

  /**
   * Tests not allowing users to disable an access policy on a node.
   */
  public function testDoNotRemoveAccessPolicyOnEntityWhenNotAllowed() {
    // Allow the access policy to be disabled.
    \Drupal::configFactory()->getEditable('access_policy.settings')
      ->set('allow_empty', FALSE)
      ->save();

    // Create a published page. Under normal circumstances this will always
    // be visible.
    $node = $this->drupalCreateNode([
      'type' => 'page',
      'status' => 1,
    ]);

    // Assign test_policy access policy to the node.
    $access_policy = \Drupal::entityTypeManager()->getStorage('access_policy')->load('test_policy');
    $this->contentAccessPolicyManager->assign($node, $access_policy);

    // Having edit any page content and edit test_policy content should give
    // you access to the content.
    $web_user = $this->drupalCreateUser([
      'set entity access policy',
      'assign test_policy access policy',
    ]);
    $this->drupalLogin($web_user);

    $this->drupalGet('node/' . $node->id() . '/access');
    $this->assertSession()->statusCodeEquals(200);

    // Assert that the "Disabled" option exists in the select list.
    $this->assertSession()->pageTextNotContains('- None -');

    // Attempt to remove the access policy. This should not be allowed.
    try {
      $edit = [
        'access_policy' => '',
      ];
      $this->submitForm($edit, 'Save');
    }
    catch (\InvalidArgumentException $e) {
      // Expected exception; just continue testing.
    }

    $node = $this->drupalGetNodeByTitle($node->label());
    $empty = $this->contentAccessPolicyManager->isEmpty($node);

    $this->assertFalse($empty, 'The access policy should not have been removed.');
    $this->assertSession()->fieldValueEquals('access_policy', 'test_policy');
  }

  /**
   * Tests assigning an empty policy if no applicable policy was found.
   */
  public function testAssignEmptyPolicyIfNoneFound() {
    // Even if allow empty is FALSE it should still set it to empty if no
    // applicable policy was found.
    \Drupal::configFactory()->getEditable('access_policy.settings')
      ->set('allow_empty', FALSE)
      ->save();

    // This user can't assign any access policies.
    $web_user = $this->drupalCreateUser([
      'set entity access policy',
    ]);
    $this->drupalLogin($web_user);

    $node = $this->drupalCreateNode([
      'type' => 'page',
      'status' => 1,
    ]);

    // Confirm that the policy was not assigned.
    $node = $this->drupalGetNodeByTitle($node->label());
    $empty = $this->contentAccessPolicyManager->isEmpty($node);
    $this->assertTrue($empty, 'The access policy should not have been set.');

    // View the Access policy selection page as super admin and confirm that
    // '- None -' is selected and that the field is enabled.
    $web_user = $this->drupalCreateUser([
      'set entity access policy',
      'assign test_policy access policy',
    ]);

    $this->drupalLogin($web_user);
    $this->drupalGet('node/' . $node->id() . '/access');
    $this->assertSession()->statusCodeEquals(200);

    // Assert that the "- None -" option exists in the select list.
    $this->assertSession()->pageTextContains('- None -');
    $this->assertSession()->fieldEnabled('access_policy');

    // Change the policy to test_policy. This should then remove '- None -'.
    $edit = [
      'access_policy' => 'test_policy',
    ];
    $this->submitForm($edit, 'Save');
    $this->assertSession()->pageTextNotContains('- None -');
  }

}
