<?php

namespace Drupal\Tests\access_policy\Functional;

use Drupal\access_policy\Entity\AccessPolicy;

/**
 * Ensures that access rule validation works.
 *
 * @group access_policy
 */
class AccessRulesEntityPermissionsTest extends AccessPolicyTestBase {

  /**
   * Modules to enable.
   *
   * @var array
   */
  protected static $modules = [
    'access_policy',
    'access_policy_test',
    'filter',
    'node',
    'field',
    'user',
    'datetime',
  ];

  /**
   * {@inheritdoc}
   */
  protected $defaultTheme = 'stark';

  /**
   * Set to TRUE to strict check all configuration saved.
   *
   * @var bool
   *
   * @see \Drupal\Core\Config\Development\ConfigSchemaChecker
   */
  protected $strictConfigSchema = FALSE;

  /**
   * {@inheritdoc}
   */
  public function setUp(): void {
    parent::setup();

    $policy = AccessPolicy::create([
      'id' => 'test_policy_with_rules',
      'label' => 'Test policy with rules',
      'target_entity_type_id' => 'node',
    ]);
    $policy->save();
    $access_rule = $this->accessRuleManager->getHandler('node', 'matching_text');
    $policy->addAccessRule($access_rule);
    $policy->save();
  }

  /**
   * Tests view permissions restricted by access policy with access rules.
   */
  public function testViewEntityWithAccessRules() {
    // Create a node with a value for field_text. This will be validated by
    // the access rule plugin.
    $node = $this->drupalCreateNode([
      'type' => 'page',
      'status' => 1,
      'field_text' => 'Matching text',
    ]);

    // Assign an access policy to the node. This now adds in a restriction
    // to users.
    $this->assignAccessPolicy($node, 'test_policy_with_rules');

    // Attempt to view the content that has matching text with the user. If the
    // text matches then they should be allowed.
    $web_user_valid = $this->drupalCreateUser([
      'access content',
      'access administration pages',
      'access content overview',
      'administer nodes',
      'view test_policy_with_rules content',
    ]);
    $web_user_valid->set('field_text', 'Matching text');
    $web_user_valid->save();
    $this->drupalLogin($web_user_valid);

    $this->drupalGet('node/' . $node->id());
    $this->assertSession()->statusCodeEquals(200);

    $this->drupalGet('admin/content');
    $this->assertSession()->statusCodeEquals(200);
    $this->assertSession()->pageTextContainsOnce($node->label());
    $this->drupalLogout();

    // Create a user without any value for field_text. They should not be able
    // to see the node.
    $web_user_invalid = $this->drupalCreateUser([
      'access content',
      'access administration pages',
      'access content overview',
      'administer nodes',
      'view test_policy_with_rules content',
    ]);
    $this->drupalLogin($web_user_invalid);

    $this->entityTypeManager->getStorage('node')->resetCache();
    $this->drupalGet('node/' . $node->id());
    $this->assertSession()->statusCodeEquals(403);

    $this->drupalGet('admin/content');
    $this->assertSession()->statusCodeEquals(200);
    $this->assertSession()->pageTextNotContains($node->label());
  }

  /**
   * Tests view permissions restricted by access policy with multiple
   * access rules evaluating as AND.
   */
  public function testViewEntityWithMultipleAndAccessRules() {

    $policy = AccessPolicy::create([
      'id' => 'test_policy_with_and_rules',
      'label' => 'Test policy with multiple AND rules',
      'access_rule_operator' => 'AND',
      'target_entity_type_id' => 'node',
    ]);
    $policy->save();
    $policy->addAccessRule($this->accessRuleManager->getHandler('node', 'matching_text'));
    $policy->addAccessRule($this->accessRuleManager->getHandler('node', 'is_own'));
    $policy->save();

    // Create a user with the right permissions and matching text.
    $web_user = $this->drupalCreateUser([
      'access content',
      'access administration pages',
      'access content overview',
      'view test_policy_with_and_rules content',
    ]);
    $web_user->set('field_text', 'Matching text');
    $web_user->save();
    $this->drupalLogin($web_user);

    // Attempt to view the content that has matching text with the user. If the
    // text matches AND the content is published then they should be allowed.
    $node = $this->drupalCreateNode([
      'type' => 'page',
      'uid' => $web_user->id(),
      'status' => 1,
      'field_text' => 'Matching text',
    ]);
    $this->assignAccessPolicy($node, 'test_policy_with_and_rules');

    $this->drupalGet('node/' . $node->id());
    $this->assertSession()->statusCodeEquals(200);

    // Create a node with a different uid and matching value for field_text.
    // this should fail validation.
    $diff_author = $this->createUser();
    $node_diff_owner = $this->drupalCreateNode([
      'type' => 'page',
      'status' => 1,
      'uid' => $diff_author->id(),
      'field_text' => 'Matching text',
    ]);
    $this->assignAccessPolicy($node_diff_owner, 'test_policy_with_and_rules');

    $this->drupalGet('node/' . $node_diff_owner->id());
    $this->assertSession()->statusCodeEquals(403);

    $this->drupalGet('admin/content');
    $this->assertSession()->statusCodeEquals(200);
    $this->assertSession()->pageTextContainsOnce($node->label());
    $this->assertSession()->pageTextNotContains($node_diff_owner->label());
  }

  /**
   * Tests view permissions restricted by access policy with multiple
   * access rules evaluating as OR.
   */
  public function testViewEntityWithMultipleOrAccessRules() {

    $policy = AccessPolicy::create([
      'id' => 'test_policy_with_or_rules',
      'label' => 'Test policy with OR rules',
      'access_rule_operator' => 'OR',
      'target_entity_type_id' => 'node',
    ]);
    $policy->addAccessRule($this->accessRuleManager->getHandler('node', 'matching_text'));
    $policy->addAccessRule($this->accessRuleManager->getHandler('node', 'is_own'));
    $policy->save();
    // Create a user with the right permissions and matching text.
    $web_user = $this->drupalCreateUser([
      'access content',
      'access administration pages',
      'access content overview',
      'administer nodes',
      'view test_policy_with_or_rules content',
    ]);
    $web_user->set('field_text', 'Matching text');
    $web_user->save();
    $this->drupalLogin($web_user);

    // Attempt to view the content that has matching text with the user. If the
    // text matches OR the content is published then they should be allowed.
    $node = $this->drupalCreateNode([
      'type' => 'page',
      'uid' => $web_user->id(),
      'status' => 1,
      'field_text' => 'Matching text',
    ]);
    $this->assignAccessPolicy($node, 'test_policy_with_or_rules');

    $this->drupalGet('node/' . $node->id());
    $this->assertSession()->statusCodeEquals(200);

    // Create a node with a different uid and matching value for field_text.
    // this should pass validation because one of them is true.
    $node_diff_owner = $this->drupalCreateNode([
      'type' => 'page',
      'status' => 1,
      'uid' => 100,
      'field_text' => 'Matching text',
    ]);
    $this->assignAccessPolicy($node_diff_owner, 'test_policy_with_or_rules');

    $this->drupalGet('node/' . $node_diff_owner->id());
    $this->assertSession()->statusCodeEquals(200);

    // Create a node with a different uid and different value for field_text.
    // This should not pass validation because none of them are true.
    $node_no_match = $this->drupalCreateNode([
      'type' => 'page',
      'status' => 1,
      'uid' => 100,
      'field_text' => 'Different text',
    ]);
    $this->assignAccessPolicy($node_no_match, 'test_policy_with_or_rules');
    $this->drupalGet('node/' . $node_no_match->id());
    $this->assertSession()->statusCodeEquals(403);

    $this->drupalGet('admin/content');
    $this->assertSession()->statusCodeEquals(200);
    $this->assertSession()->pageTextContainsOnce($node->label());
    $this->assertSession()->pageTextContainsOnce($node_diff_owner->label());
    $this->assertSession()->pageTextNotContains($node_no_match->label());

  }

  /**
   * Tests bypassed access rules.
   */
  public function testViewEntityWithBypassedAccessRules() {
    // Create a node with a value for field_text. This will be validated by
    // the access rule plugin.
    $node = $this->drupalCreateNode([
      'type' => 'page',
      'status' => 1,
      'field_text' => 'Matching text',
    ]);

    $this->assignAccessPolicy($node, 'test_policy_with_rules');

    // Create a user without any value for field_text but they have the bypass
    // permission. They should be able to see the node.
    $web_user_invalid = $this->drupalCreateUser([
      'access content',
      'access administration pages',
      'access content overview',
      'administer nodes',
      'view test_policy_with_rules content',
      'bypass test_policy_with_rules access rules',
    ]);
    $this->drupalLogin($web_user_invalid);

    $this->drupalGet('node/' . $node->id());
    $this->assertSession()->statusCodeEquals(200);

    $this->drupalGet('admin/content');
    $this->assertSession()->statusCodeEquals(200);
    $this->assertSession()->pageTextContainsOnce($node->label());
  }

}
