<?php

namespace Drupal\Tests\access_policy\Kernel;

use Drupal\access_policy\AccessPolicyQueryAlter;
use Drupal\Core\Database\Database;
use Drupal\Core\Entity\EntityInterface;
use Drupal\field\Entity\FieldConfig;
use Drupal\field\Entity\FieldStorageConfig;
use Drupal\KernelTests\KernelTestBase;
use Drupal\Tests\node\Traits\NodeCreationTrait;
use Drupal\Tests\user\Traits\UserCreationTrait;

/**
 * Access policy kernel test base.
 */
abstract class AccessPolicyKernelTestBase extends KernelTestBase {

  use NodeCreationTrait;
  use UserCreationTrait;

  /**
   * Set to TRUE to strict check all configuration saved.
   *
   * @var bool
   *
   * @see \Drupal\Core\Config\Development\ConfigSchemaChecker
   */
  protected $strictConfigSchema = FALSE;

  /**
   * The database connection.
   *
   * @var \Drupal\Core\Database\Connection
   */
  protected $connection;

  /**
   * The entity type manager.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * The content access policy manager.
   *
   * @var \Drupal\access_policy\ContentAccessPolicyManager
   */
  protected $contentAccessPolicyManager;

  /**
   * The access rule manager.
   *
   * @var \Drupal\access_policy\AccessPolicyHandlerManager
   */
  protected $accessRuleManager;

  /**
   * The access policy validator.
   *
   * @var \Drupal\access_policy\AccessPolicyValidatorInterface
   */
  protected $accessPolicyValidator;

  /**
   * The modules required for this test.
   *
   * @var string[]
   */
  protected static $modules = [
    'field',
    'filter',
    'node',
    'options',
    'access_policy',
    'access_policy_test',
    'system',
    'text',
    'user',
  ];

  /**
   * {@inheritdoc}
   *
   * @throws \Exception
   */
  protected function setUp(): void {
    parent::setUp();
    $this->installConfig(['system']);
    $this->installConfig(['node']);
    $this->installConfig(['filter']);
    $this->installSchema('system', ['sequences']);
    $this->installSchema('node', ['node_access']);
    $this->installEntitySchema('user');
    $this->installEntitySchema('node');
    $this->installEntitySchema('access_policy');
    // @todo remove in beta1.
    $this->installEntitySchema('content_access_policy');

    $this->connection = Database::getConnection();
    $this->entityTypeManager = $this->container->get('entity_type.manager');
    $this->contentAccessPolicyManager = \Drupal::service('access_policy.content_policy_manager');
    $this->accessRuleManager = $this->container->get('plugin.manager.access_policy.access_rule');
    $this->accessPolicyValidator = $this->container->get('access_policy.validator');
  }

  /**
   * Get the access policy query alter class.
   *
   * @return \Drupal\access_policy\AccessPolicyQueryAlter
   *   The access policy query alter handler.
   */
  protected function getQueryAlter() {
    return \Drupal::service('class_resolver')
      ->getInstanceFromDefinition(AccessPolicyQueryAlter::class);
  }

  /**
   * Create new node type.
   *
   * @param string $type
   *   The node type.
   * @param string $name
   *   The type name.
   */
  public function createNodeType(string $type, string $name) {
    $this->entityTypeManager->getStorage('node_type')->create([
      'type' => $type,
      'name' => $name,
    ])->save();
  }

  /**
   * Create a node and assign values.
   *
   * @param array $node_properties
   *   The node properties.
   *
   * @return \Drupal\node\NodeInterface
   *   The node.
   *
   * @throws \Drupal\Core\Entity\EntityStorageException
   */
  public function createStubNode(array $node_properties = []) {
    $node = $this->createNode($node_properties);
    $node->save();
    $this->entityTypeManager->getStorage('node')->resetCache();

    return $node;
  }

  /**
   * Create a user and assign values.
   *
   * @param array $permissions
   *   The user permissions.
   * @param array $user_values
   *   The values to set.
   *
   * @return \Drupal\user\Entity\User|false
   *   The user.
   *
   * @throws \Drupal\Core\Entity\EntityStorageException
   */
  public function createStubUser(array $permissions = [], array $user_values = []) {
    $uid = rand(2, 999);
    $user = $this->createUser($permissions, $this->randomString(8), FALSE, [
      'uid' => $uid,
      'activation_status' => 'Active',
    ] + $user_values);
    $user->save();
    $this->entityTypeManager->getStorage('user')->resetCache();

    return $user;
  }

  /**
   * Create a new text field.
   *
   * @param string $field_machine_name
   *   The machine name for the field.
   * @param array $settings
   *   The field settings.
   *
   * @throws \Drupal\Core\Entity\EntityStorageException
   */
  public function createTextField(string $field_machine_name, array $settings) {
    $field_storage = FieldStorageConfig::loadByName($settings['entity_type'], $field_machine_name);
    if (!$field_storage) {
      $field_storage = FieldStorageConfig::create([
        'field_name' => $field_machine_name,
        'entity_type' => $settings['entity_type'],
        'type' => 'string',
        'cardinality' => 1,
      ]);
      $field_storage->save();
    }

    $field = FieldConfig::loadByName($settings['entity_type'], $settings['bundle'], $field_machine_name);
    if (!$field) {
      $field = FieldConfig::create([
        'field_storage' => $field_storage,
        'field_name' => 'sales_org_id',
        'bundle' => 'user',
        'translatable' => FALSE,
      ]);
      $field->save();
    }
  }

  /**
   * Create a new entity reference field.
   *
   * @param string $field_machine_name
   *   The machine name for the field.
   * @param array $settings
   *   The field settings.
   *
   * @throws \Drupal\Core\Entity\EntityStorageException
   */
  public function createEntityReferenceField(string $field_machine_name, array $settings) {
    $field_storage = FieldStorageConfig::loadByName($settings['entity_type'], $field_machine_name);
    if (!$field_storage) {
      $field_storage = FieldStorageConfig::create([
        'field_name' => $field_machine_name,
        'entity_type' => $settings['entity_type'],
        'type' => 'entity_reference',
        'cardinality' => 1,
        'settings' => [
          'target_type' => $settings['target_type'],
        ],
      ]);
      $field_storage->save();
    }

    $field = FieldConfig::loadByName($settings['entity_type'], $settings['bundle'], $field_machine_name);
    if (!$field) {
      $field = FieldConfig::create([
        'field_storage' => $field_storage,
        'field_name' => $field_machine_name,
        'bundle' => $settings['bundle'],
        'settings' => [
          'handler' => 'default:taxonomy_term',
          'handler_settings' => [
            'target_bundles' => $settings['target_bundles'],
            'sort' => [
              'field' => 'name',
              'direction' => 'asc',
            ],
            'auto_create' => TRUE,
            'auto_create_bundle' => '',
          ],
        ],
        'translatable' => FALSE,
      ]);
      $field->save();
    }
  }

  /**
   * Helper function for assigning an access policy to a node.
   *
   * @param \Drupal\Core\Entity\EntityInterface $entity
   *   The entity.
   * @param mixed $access_policy
   *   The access policy.
   *
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   */
  protected function assignAccessPolicy(EntityInterface $entity, $access_policy) {
    if (!is_array($access_policy)) {
      $access_policy = [$access_policy];
    }
    $entity->set('access_policy', $access_policy);
    $entity->save();
  }

  /**
   * Create access policy entity.
   *
   * @param string $name
   *   The name of the policy.
   * @param array $values
   *   The policy vale.
   *
   * @return \Drupal\access_policy\Entity\AccessPolicyInterface
   *   The created access policy.
   */
  protected function createAccessPolicy($name, array $values = []) {
    $values += [
      'id' => $name,
      'label' => $name,
    ];
    $access_policy = $this->entityTypeManager->getStorage('access_policy')
      ->create($values);
    $access_policy->save();

    return $access_policy;
  }

  /**
   * Assert query results.
   *
   * @param string $entity_type
   *   The entity type.
   * @param mixed $expected
   *   The expected results.
   */
  protected function assertQueryResults($entity_type, $expected) {
    // @todo add support for other entity types.
    $entity_type = $this->entityTypeManager->getDefinition($entity_type);
    $query = $this->connection->select('node_field_data')
      ->fields('node_field_data', ['nid']);

    $this->getQueryAlter()->queryAlter($query, $entity_type);
    $result = $query->execute()->fetchCol();
    $this->assertEquals($expected, $result, 'The query did not return the right entities.');
  }

}
