<?php

namespace Drupal\Tests\access_policy\Kernel;

use Drupal\node\Entity\NodeType;
use Drupal\access_policy\Entity\AccessPolicy;

/**
 * Tests Access Policy storage.
 *
 * @group access_policy
 */
class EntityAccessPolicyStorageTest extends AccessPolicyKernelTestBase {

  /**
   * {@inheritdoc}
   *
   * @throws \Exception
   */
  protected function setUp(): void {
    parent::setUp();

    $this->selectionRuleManager = \Drupal::service('plugin.manager.access_policy.selection_rule');

    AccessPolicy::create([
      'id' => 'policy_test_1',
      'label' => 'Policy test 1',
      'access_rules' => [],
      'target_entity_type_id' => 'node',
      'selection_set' => ['restricted'],
      'weight' => 0,
    ])->save();

    AccessPolicy::create([
      'id' => 'policy_test_2',
      'label' => 'Policy test 2',
      'access_rules' => [],
      'target_entity_type_id' => 'node',
      'selection_set' => ['restricted'],
      'weight' => -10,
    ])->save();

    NodeType::create(['type' => 'page', 'name' => 'Page']);
    NodeType::create(['type' => 'article', 'name' => 'Article']);
  }

  /**
   * Tests saving and deleting an access policy on an entity.
   *
   * @throws \Exception
   */
  public function testSaveAndDeleteAccessPolicyOnEntity() {
    $entity_type = $this->entityTypeManager->getDefinition('node');
    $storage = \Drupal::service('access_policy.storage.entity');
    // Make sure that the table is installed.
    $storage->installSchema($entity_type);

    $node = $this->createNode([
      'type' => 'page',
      'title' => 'Page',
    ]);
    $node->save();

    $storage->write($node, 'policy_test_1');

    // Write a single access policy.
    $policy = $storage->read($node);
    $this->assertEquals(['policy_test_1'], $policy, 'The policy should have been saved.');

    // Write multiple access policies to the node.
    $storage->write($node, ['policy_test_1', 'policy_test_2']);
    $policy = $storage->read($node);
    $this->assertEquals(['policy_test_1', 'policy_test_2'], $policy, 'Two policies should have been saved.');

    $storage->delete($node);
    $policy = $storage->read($node);
    $this->assertEmpty($policy, 'The policy should have been deleted.');
  }

  /**
   * This will attempt to save an invalid access policy on the entity.
   */
  public function testSaveInvalidAccessPolicyOnEntity() {
    $entity_type = $this->entityTypeManager->getDefinition('node');
    $storage = \Drupal::service('access_policy.storage.entity');
    // Make sure that the table is installed.
    $storage->installSchema($entity_type);

    $node = $this->createNode([
      'type' => 'article',
      'title' => 'Article',
    ]);
    $node->save();

    // policy_test_2 is supported.
    $storage->write($node, ['policy_test_2']);

    // Try to add 'policy_test_3', it should fail and not change anything.
    try {
      $storage->write($node, ['policy_test_3', 'policy_test_2']);
    }
    catch (\Exception $e) {
      $this->assertEquals('The "policy_test_3" access policy does not exist.', $e->getMessage());
    }
    $policy = $storage->read($node);
    $this->assertEquals(['policy_test_2'], $policy, 'The policy should not have changed.');
  }

}
