<?php

namespace Drupal\Tests\access_policy\Kernel\Plugin\AccessRule;

use Drupal\access_policy\Entity\AccessPolicy;
use Drupal\access_policy\Plugin\OperatorValidationTrait;
use Drupal\field\Entity\FieldConfig;
use Drupal\field\Entity\FieldStorageConfig;
use Drupal\Tests\taxonomy\Traits\TaxonomyTestTrait;
use Drupal\Tests\access_policy\Kernel\AccessPolicyKernelTestBase;

/**
 * Tests EntityField base plugin features.
 *
 * @group access_policy
 */
class EntityFieldNumericTest extends AccessPolicyKernelTestBase {

  use TaxonomyTestTrait;
  use OperatorValidationTrait;

  /**
   * The modules to install.
   *
   * @var string[]
   */
  protected static $modules = [
    'field',
    'filter',
    'node',
    'options',
    'access_policy',
    'access_policy_test',
    'system',
    'text',
    'user',
    'datetime',
  ];

  /**
   * {@inheritdoc}
   */
  protected function setUp(): void {
    parent::setUp();

    $this->installConfig(['access_policy_test']);

    // Add an integer field to the node.
    $field_storage = FieldStorageConfig::create([
      'field_name' => 'field_integer',
      'entity_type' => 'node',
      'type' => 'integer',
      'cardinality' => -1,
    ]);
    $field_storage->save();
    $field = FieldConfig::create([
      'field_storage' => $field_storage,
      'field_name' => 'field_integer',
      'bundle' => 'page',
      'translatable' => FALSE,
    ]);
    $field->save();
  }

  /**
   * Tests the numeric entity field access rule.
   */
  public function testEntityFieldNumeric() {

    $policy = AccessPolicy::create([
      'id' => 'numeric',
      'label' => 'Access policy with numeric values',
      'access_rules' => [],
      'target_entity_type_id' => 'node',
    ]);
    $policy->save();

    // Add a greater than operator.
    $access_rule = $this->accessRuleManager->getHandler('node', 'field_integer');
    $access_rule->setSettings([
      'value' => 50,
      'operator' => '>',
      'query' => TRUE,
    ]);
    $policy->addAccessRule($access_rule);
    $policy->save();

    // 75 is between 50 and 100 and should be visible.
    $node_1 = $this->createNode([
      'title' => 'Node 1 (With int: 75)',
      'type' => 'page',
      'status' => 1,
      'field_integer' => [
        75,
      ],
    ]);

    // Node 2 is less than 50 and should not be visible.
    $node_2 = $this->createNode([
      'title' => 'Node 2 (With int: 25)',
      'type' => 'page',
      'status' => 1,
      'field_integer' => [
        25,
      ],
    ]);

    $this->assignAccessPolicy($node_1, 'numeric');
    $this->assignAccessPolicy($node_2, 'numeric');

    $admin_user = $this->createUser([
      'access content',
      'view numeric content',
    ]);
    $admin_user->save();

    $this->setCurrentUser($admin_user);

    $access = $this->accessPolicyValidator->validate($node_1, $admin_user, 'view');
    $this->assertTrue($access, 'The user should have access.');

    $access = $this->accessPolicyValidator->validate($node_2, $admin_user, 'view');
    $this->assertFalse($access, 'The user should not have access.');

    $this->assertQueryResults('node', [
      $node_1->id(),
    ]);

    // Update the access rule and run validation again.
    $access_rule_settings = $access_rule->getSettings();
    $access_rule_settings['operator'] = '<';
    $policy->updateAccessRule('field_integer', $access_rule_settings);
    $policy->save();

    $access = $this->accessPolicyValidator->validate($node_1, $admin_user, 'view');
    $this->assertFalse($access, 'The user should not have access.');

    $access = $this->accessPolicyValidator->validate($node_2, $admin_user, 'view');
    $this->assertTrue($access, 'The user should have access.');

    $this->assertQueryResults('node', [
      $node_2->id(),
    ]);
  }

  /**
   * Tests a node with numeric field values between values.
   */
  public function testEntityFieldNumericWithMultipleRules() {
    $policy = AccessPolicy::create([
      'id' => 'integer_between',
      'label' => 'Access policy with between values.',
      'access_rules' => [],
      'target_entity_type_id' => 'node',
    ]);
    $policy->save();

    // Add a less than operator.
    $access_rule = $this->accessRuleManager->getHandler('node', 'field_integer');
    $access_rule->setSettings([
      'value' => 100,
      'operator' => '<',
      'query' => TRUE,
    ]);
    $policy->addAccessRule($access_rule);
    $policy->save();

    // Add a greater than operator.
    $access_rule = $this->accessRuleManager->getHandler('node', 'field_integer');
    $access_rule->setSettings([
      'value' => 50,
      'operator' => '>',
      'query' => TRUE,
    ]);
    $policy->addAccessRule($access_rule);
    $policy->save();

    // Add a greater than operator.
    $access_rule = $this->accessRuleManager->getHandler('node', 'field_integer');
    $access_rule->setSettings([
      'value' => 101,
      'operator' => '!=',
      'query' => TRUE,
    ]);
    $policy->addAccessRule($access_rule);
    $policy->save();

    // 75 is between 50 and 100 and should be visible.
    $node_1 = $this->createNode([
      'title' => 'Node 1 (With int: 75)',
      'type' => 'page',
      'status' => 1,
      'field_integer' => [
        75,
      ],
    ]);

    // 25 is less than 50 and should not be visible.
    $node_2 = $this->createNode([
      'title' => 'Node 2 (With int: 25)',
      'type' => 'page',
      'status' => 1,
      'field_integer' => [
        25,
      ],
    ]);

    // 125 is greater than 100 and should not be visible.
    $node_3 = $this->createNode([
      'title' => 'Node 3 (With int: 125)',
      'type' => 'page',
      'status' => 1,
      'field_integer' => [
        125,
      ],
    ]);

    // 101 is not allowed and should not be visible.
    $node_4 = $this->createNode([
      'title' => 'Node 4 (With int: 101)',
      'type' => 'page',
      'status' => 1,
      'field_integer' => [
        101,
      ],
    ]);

    $this->assignAccessPolicy($node_1, 'integer_between');
    $this->assignAccessPolicy($node_2, 'integer_between');
    $this->assignAccessPolicy($node_3, 'integer_between');
    $this->assignAccessPolicy($node_4, 'integer_between');

    // We're not comparing field values with a user so it's not necessary to set
    // it.
    $admin_user = $this->createUser([
      'access content',
      'access administration pages',
      'access content overview',
      'administer nodes',
      'view integer_between content',
    ]);
    $admin_user->save();

    $access = $this->accessPolicyValidator->validate($node_1, $admin_user, 'view');
    $this->assertTrue($access, 'The user should have access.');

    $access = $this->accessPolicyValidator->validate($node_2, $admin_user, 'view');
    $this->assertFalse($access, 'The user should not have access.');

    $access = $this->accessPolicyValidator->validate($node_3, $admin_user, 'view');
    $this->assertFalse($access, 'The user should not have access.');

    $access = $this->accessPolicyValidator->validate($node_4, $admin_user, 'view');
    $this->assertFalse($access, 'The user should not have access.');

    $this->setCurrentUser($admin_user);
    $this->assertQueryResults('node', [
      $node_1->id(),
    ]);

  }

}
