<?php

namespace Drupal\Tests\access_policy\Kernel\Plugin\AccessRule;

use Drupal\access_policy\Entity\AccessPolicy;
use Drupal\access_policy\Plugin\OperatorValidationTrait;
use Drupal\Tests\taxonomy\Traits\TaxonomyTestTrait;
use Drupal\Tests\access_policy\Kernel\AccessPolicyKernelTestBase;

/**
 * Tests EntityField base plugin features.
 *
 * @group access_policy
 */
class EntityFieldStringTest extends AccessPolicyKernelTestBase {

  use TaxonomyTestTrait;
  use OperatorValidationTrait;

  /**
   * The modules to install.
   *
   * @var string[]
   */
  protected static $modules = [
    'field',
    'filter',
    'node',
    'options',
    'access_policy',
    'access_policy_test',
    'system',
    'text',
    'user',
    'datetime',
  ];

  /**
   * {@inheritdoc}
   */
  protected function setUp(): void {
    parent::setUp();

    $this->installConfig(['access_policy_test']);
  }

  /**
   * Tests entity field string equals.
   */
  public function testEntityFieldStringEquals() {
    $policy = AccessPolicy::create([
      'id' => 'string_equals',
      'label' => 'Access policy string equals',
      'access_rules' => [],
      'target_entity_type_id' => 'node',
    ]);
    $policy->save();

    // Add an equals operator.
    $access_rule = $this->accessRuleManager->getHandler('node', 'field_text');
    $access_rule->setSettings([
      'value' => 'Foo',
      'operator' => '=',
      'query' => TRUE,
    ]);
    $policy->addAccessRule($access_rule);
    $policy->save();

    // Add a not equals operator.
    $access_rule = $this->accessRuleManager->getHandler('node', 'field_text');
    $access_rule->setSettings([
      'value' => 'Bar',
      'operator' => '!=',
      'query' => TRUE,
    ]);
    $policy->addAccessRule($access_rule);
    $policy->save();

    // Node 1 has foo and should be accessible.
    $node_1 = $this->createNode([
      'title' => 'Node 1 (With string: Foo)',
      'type' => 'page',
      'status' => 1,
      'field_text' => [
        'Foo',
      ],
    ]);

    // This has Bar and therefore should not be accessible.
    $node_2 = $this->createNode([
      'title' => 'Node 2 (With string: Bar)',
      'type' => 'page',
      'status' => 1,
      'field_text' => [
        'Bar',
      ],
    ]);

    // Because this has one value that matches it should be visible.
    $node_3 = $this->createNode([
      'title' => 'Node 3 (With multiple values - allowed)',
      'type' => 'page',
      'status' => 1,
      'field_text' => [
        'Foo',
        'Bar',
      ],
    ]);

    $this->assignAccessPolicy($node_1, 'string_equals');
    $this->assignAccessPolicy($node_2, 'string_equals');
    $this->assignAccessPolicy($node_3, 'string_equals');

    $admin_user = $this->createUser([
      'access content',
      'view string_equals content',
    ]);
    $admin_user->save();

    $access = $this->accessPolicyValidator->validate($node_1, $admin_user, 'view');
    $this->assertTrue($access, 'The user should have access.');

    $access = $this->accessPolicyValidator->validate($node_2, $admin_user, 'view');
    $this->assertFalse($access, 'The user should not have access.');

    $access = $this->accessPolicyValidator->validate($node_3, $admin_user, 'view');
    $this->assertTrue($access, 'The user should have access.');

    $this->setCurrentUser($admin_user);
    $entity_type = $this->entityTypeManager->getDefinition('node');
    $query = $this->connection->select('node_field_data')
      ->fields('node_field_data', ['nid']);

    $this->getQueryAlter()->queryAlter($query, $entity_type);
    $result = $query->execute()->fetchCol();
    $expected = [
      $node_1->id(),
      $node_3->id(),
    ];
    $this->assertEquals($expected, $result, 'The query did not return the right nodes.');
  }

  /**
   * Tests entity field string not equals.
   */
  public function testEntityFieldStringNotEquals() {
    $policy = AccessPolicy::create([
      'id' => 'string_not_equals',
      'label' => 'Access policy string not equals',
      'access_rules' => [],
      'target_entity_type_id' => 'node',
    ]);
    $policy->save();

    // Add a not equals operator.
    $access_rule = $this->accessRuleManager->getHandler('node', 'field_text');
    $access_rule->setSettings([
      'value' => 'Bar',
      'operator' => '!=',
      'query' => TRUE,
    ]);
    $policy->addAccessRule($access_rule);
    $policy->save();

    // Node 1 has foo which does not equal Bar and should be accessible.
    $node_1 = $this->createNode([
      'title' => 'Node 1 (With string: Foo)',
      'type' => 'page',
      'status' => 1,
      'field_text' => [
        'Foo',
      ],
    ]);

    // This has Bar and therefore should not be accessible.
    $node_2 = $this->createNode([
      'title' => 'Node 2 (With string: Bar)',
      'type' => 'page',
      'status' => 1,
      'field_text' => [
        'Bar',
      ],
    ]);

    $this->assignAccessPolicy($node_1, 'string_not_equals');
    $this->assignAccessPolicy($node_2, 'string_not_equals');

    // We're not comparing field values with a user so it's not necessary to set
    // it.
    $admin_user = $this->createUser([
      'access content',
      'view string_not_equals content',
    ]);
    $admin_user->save();

    $access = $this->accessPolicyValidator->validate($node_1, $admin_user, 'view');
    $this->assertTrue($access, 'The user should have access.');

    $access = $this->accessPolicyValidator->validate($node_2, $admin_user, 'view');
    $this->assertFalse($access, 'The user should not have access.');

    $this->setCurrentUser($admin_user);
    $this->assertQueryResults('node', [
      $node_1->id(),
    ]);
  }

  /**
   * Tests entity field string contains and not contains.
   */
  public function testEntityFieldContainsAndNotContains() {
    $policy = AccessPolicy::create([
      'id' => 'string_contains',
      'label' => 'Access policy contains',
      'access_rules' => [],
      'target_entity_type_id' => 'node',
    ]);
    $policy->save();

    // Add a contains operator.
    $access_rule = $this->accessRuleManager->getHandler('node', 'field_text');
    $access_rule->setSettings([
      'value' => 'Foo',
      'operator' => 'contains',
      'query' => TRUE,
    ]);
    $policy->addAccessRule($access_rule);
    $policy->save();

    // Add a not contains operator.
    $access_rule = $this->accessRuleManager->getHandler('node', 'field_text');
    $access_rule->setSettings([
      'value' => 'Bar',
      'operator' => 'not contains',
      'query' => TRUE,
    ]);
    $policy->addAccessRule($access_rule);
    $policy->save();

    // Node 1 has foo but doesn't have bar and should be accessible.
    $node_1 = $this->createNode([
      'title' => 'Node 1 (With string: Foo)',
      'type' => 'page',
      'status' => 1,
      'field_text' => [
        'Text with Foo',
      ],
    ]);

    $node_2 = $this->createNode([
      'title' => 'Node 2 (With string: Bar)',
      'type' => 'page',
      'status' => 1,
      'field_text' => [
        'Text with Bar',
      ],
    ]);

    $node_3 = $this->createNode([
      'title' => 'Node 3 (With string: Foo Bar)',
      'type' => 'page',
      'status' => 1,
      'field_text' => [
        'Text with Foo Bar',
      ],
    ]);

    // Because this has one value that matches it should be visible.
    $node_4 = $this->createNode([
      'title' => 'Node 4 (With multiple values - allowed)',
      'type' => 'page',
      'status' => 1,
      'field_text' => [
        'Text with Bar',
        'Text with Foo',
      ],
    ]);

    // Because this does not have any values that match it should not be
    // visible.
    $node_5 = $this->createNode([
      'title' => 'Node 5 (With multiple values - denied)',
      'type' => 'page',
      'status' => 1,
      'field_text' => [
        'Text with Bar',
        'Text with Bar again',
      ],
    ]);

    $this->assignAccessPolicy($node_1, 'string_contains');
    $this->assignAccessPolicy($node_2, 'string_contains');
    $this->assignAccessPolicy($node_3, 'string_contains');
    $this->assignAccessPolicy($node_4, 'string_contains');
    $this->assignAccessPolicy($node_5, 'string_contains');

    // We're not comparing field values with a user so it's not necessary to set
    // it.
    $admin_user = $this->createUser([
      'access content',
      'access administration pages',
      'access content overview',
      'administer nodes',
      'view string_contains content',
    ]);
    $admin_user->save();

    $access = $this->accessPolicyValidator->validate($node_1, $admin_user, 'view');
    $this->assertTrue($access, 'The user should have access.');

    $access = $this->accessPolicyValidator->validate($node_2, $admin_user, 'view');
    $this->assertFalse($access, 'The user should not have access.');

    $access = $this->accessPolicyValidator->validate($node_3, $admin_user, 'view');
    $this->assertFalse($access, 'The user should not have access.');

    $access = $this->accessPolicyValidator->validate($node_4, $admin_user, 'view');
    $this->assertTrue($access, 'The user should have access.');

    $access = $this->accessPolicyValidator->validate($node_5, $admin_user, 'view');
    $this->assertFalse($access, 'The user should not have access.');

    $this->setCurrentUser($admin_user);
    $entity_type = $this->entityTypeManager->getDefinition('node');
    $query = $this->connection->select('node_field_data')
      ->fields('node_field_data', ['nid']);

    $this->getQueryAlter()->queryAlter($query, $entity_type);
    $result = $query->execute()->fetchCol();
    $expected = [
      $node_1->id(),
      $node_4->id(),
    ];
    $this->assertEquals($expected, $result, 'The query did not return the right nodes.');
  }

  /**
   * Tests field value starts with and not starts with.
   */
  public function testEntityFieldStartsWithAndNotStartsWith() {
    $policy = AccessPolicy::create([
      'id' => 'string_starts_with',
      'label' => 'Access policy starts with',
      'access_rules' => [],
      'target_entity_type_id' => 'node',
    ]);
    $policy->save();

    $access_rule = $this->accessRuleManager->getHandler('node', 'field_text');
    $access_rule->setSettings([
      'value' => 'Foo',
      'operator' => 'starts with',
      'query' => TRUE,
    ]);

    $policy->addAccessRule($access_rule);
    $policy->save();

    $access_rule = $this->accessRuleManager->getHandler('node', 'field_text');
    $access_rule->setSettings([
      'value' => 'Bar',
      'operator' => 'not starts with',
      'query' => TRUE,
    ]);
    $policy->addAccessRule($access_rule);
    $policy->save();

    $node_1 = $this->createNode([
      'title' => 'Node 1 (With start: Foo)',
      'type' => 'page',
      'status' => 1,
      'field_text' => [
        'Foo with text',
      ],
    ]);

    $node_2 = $this->createNode([
      'title' => 'Node 2 (With start: Bar)',
      'type' => 'page',
      'status' => 1,
      'field_text' => [
        'Bar with text',
      ],
    ]);

    // Because this has one value that matches it should be visible.
    $node_3 = $this->createNode([
      'title' => 'Node 3 (With multiple values - allowed)',
      'type' => 'page',
      'status' => 1,
      'field_text' => [
        'Foo with text',
        'Bar with text',
      ],
    ]);

    // Because this does not have any values that match it should not be
    // visible.
    $node_4 = $this->createNode([
      'title' => 'Node 4 (With multiple values - denied)',
      'type' => 'page',
      'status' => 1,
      'field_text' => [
        'Text with Foo',
        'Bar with text again',
      ],
    ]);

    $this->assignAccessPolicy($node_1, 'string_starts_with');
    $this->assignAccessPolicy($node_2, 'string_starts_with');
    $this->assignAccessPolicy($node_3, 'string_starts_with');
    $this->assignAccessPolicy($node_4, 'string_starts_with');

    // We're not comparing field values with a user so it's not necessary to set
    // it.
    $admin_user = $this->createUser([
      'access content',
      'access administration pages',
      'access content overview',
      'administer nodes',
      'view string_starts_with content',
    ]);
    $admin_user->save();

    $access = $this->accessPolicyValidator->validate($node_1, $admin_user, 'view');
    $this->assertTrue($access, 'The user should have access.');

    $access = $this->accessPolicyValidator->validate($node_2, $admin_user, 'view');
    $this->assertFalse($access, 'The user should not have access.');

    $access = $this->accessPolicyValidator->validate($node_3, $admin_user, 'view');
    $this->assertTrue($access, 'The user should have access.');

    $access = $this->accessPolicyValidator->validate($node_4, $admin_user, 'view');
    $this->assertFalse($access, 'The user should not have access.');

    $this->setCurrentUser($admin_user);
    $entity_type = $this->entityTypeManager->getDefinition('node');
    $query = $this->connection->select('node_field_data')
      ->fields('node_field_data', ['nid']);

    $this->getQueryAlter()->queryAlter($query, $entity_type);
    $result = $query->execute()->fetchCol();
    $expected = [
      $node_1->id(),
      $node_3->id(),
    ];
    $this->assertEquals($expected, $result, 'The query did not return the right nodes.');
  }

  /**
   * Tests field value ends with and not ends with.
   */
  public function testEntityFieldEndsWithAndNotEndsWith() {
    $policy = AccessPolicy::create([
      'id' => 'string_ends_with',
      'label' => 'Access policy ends with',
      'target_entity_type_id' => 'node',
    ]);
    $policy->save();

    $access_rule = $this->accessRuleManager->getHandler('node', 'field_text');
    $access_rule->setSettings([
      'value' => 'Foo',
      'operator' => 'ends with',
      'query' => TRUE,
    ]);
    $policy->addAccessRule($access_rule);
    $policy->save();

    $access_rule = $this->accessRuleManager->getHandler('node', 'field_text');
    $access_rule->setSettings([
      'value' => 'Bar',
      'operator' => 'not ends with',
      'query' => TRUE,
    ]);
    $policy->addAccessRule($access_rule);
    $policy->save();

    $node_1 = $this->createNode([
      'title' => 'Node 1 (With end: Foo)',
      'type' => 'page',
      'status' => 1,
      'field_text' => [
        'Text with Foo',
      ],
    ]);

    $node_2 = $this->createNode([
      'title' => 'Node 2 (With end: Bar)',
      'type' => 'page',
      'status' => 1,
      'field_text' => [
        'Text with Bar',
      ],
    ]);

    // Because this has one value that matches it should be visible.
    $node_3 = $this->createNode([
      'title' => 'Node 3 (With multiple values - allowed)',
      'type' => 'page',
      'status' => 1,
      'field_text' => [
        'Text with Foo',
        'Text with Bar',
      ],
    ]);

    // Because this does not have any values that match it should not be
    // visible.
    $node_4 = $this->createNode([
      'title' => 'Node 4 (With multiple values - denied)',
      'type' => 'page',
      'status' => 1,
      'field_text' => [
        'Foo with text',
        'Bar with text again',
      ],
    ]);

    $this->assignAccessPolicy($node_1, 'string_ends_with');
    $this->assignAccessPolicy($node_2, 'string_ends_with');
    $this->assignAccessPolicy($node_3, 'string_ends_with');
    $this->assignAccessPolicy($node_4, 'string_ends_with');

    // We're not comparing field values with a user so it's not necessary to set
    // it.
    $admin_user = $this->createUser([
      'access content',
      'access administration pages',
      'access content overview',
      'administer nodes',
      'view string_ends_with content',
    ]);
    $admin_user->save();

    $access = $this->accessPolicyValidator->validate($node_1, $admin_user, 'view');
    $this->assertTrue($access, 'The user should have access.');

    $access = $this->accessPolicyValidator->validate($node_2, $admin_user, 'view');
    $this->assertFalse($access, 'The user should not have access.');

    $access = $this->accessPolicyValidator->validate($node_3, $admin_user, 'view');
    $this->assertTrue($access, 'The user should have access.');

    $access = $this->accessPolicyValidator->validate($node_4, $admin_user, 'view');
    $this->assertFalse($access, 'The user should not have access.');

    $this->setCurrentUser($admin_user);
    $this->assertQueryResults('node', [
      $node_1->id(),
      $node_3->id(),
    ]);
  }

}
